from __future__ import annotations

import contextlib
import copy
from collections.abc import AsyncGenerator, AsyncIterator
from datetime import UTC, datetime, timedelta
from typing import Any, Optional, Self, override

from yarl import URL

from ...converter import converter
from ...types.info import RepositoryInfo
from ...types.records import Record, RecordId, RecordList
from ...types.requests import Request, RequestType
from ...types.rest import RESTHits, RESTPaginationLinks
from ..base_client import AsyncRecordsClient, RecordStatus
from ..connection import AsyncConnection
from .requests import AsyncInvenioRequestsClient

OPENSEARCH_SCAN_WINDOW = 5000
OPENSEARCH_SCAN_PAGE = 100


class AsyncInvenioRecordsClient(AsyncRecordsClient):
    def __init__(
        self,
        connection: AsyncConnection,
        info: RepositoryInfo,
        requests_client: AsyncInvenioRequestsClient,
    ):
        """Initialize the client."""
        self._connection = connection
        self._info = info
        self._model: str | None = None
        self._status: RecordStatus = RecordStatus.ALL
        self._requests_client = requests_client

    @override
    def with_model(self, model: str) -> Self:
        """Return a new client limited to the given model."""
        ret = copy.copy(self)
        ret._model = model
        return ret

    @override
    @property
    def published_records(self) -> Self:
        """Return a new client limited to published records."""
        ret = copy.copy(self)
        ret._status = RecordStatus.PUBLISHED
        return ret

    @override
    @property
    def draft_records(self) -> Self:
        """Return a new client limited to draft records."""
        ret = copy.copy(self)
        ret._status = RecordStatus.DRAFT
        return ret

    @override
    async def create(
        self,
        data: dict[str, Any],
        *,
        model: str | None = None,
        community: str | None = None,
        workflow: str | None = None,
        idempotent: bool = False,
        files_enabled: bool = True,
    ) -> Record:
        """Create a new record in the repository.

        :param data:            the metadata of the record
        :param community:       community in which the record should be created
        :param workflow:        the workflow to use for the record, if not provided
                                the default workflow of the community is used
        :param idempotent:      if True, the operation is idempotent and can be retried on network errors.
                                Use only if you know that the operation is idempotent, for example that
                                you use PID generator that takes the persistent identifier from the data.
        :return:                the created record
        """
        if idempotent:
            raise NotImplementedError("Idempotent for create not implemented yet")

        # if model is not provided, try to use currently set model
        model = model or self._model

        # if there is no model, try to estimate it by $schema in the data
        if not model and "$schema" in data:
            model_schema = data["$schema"]
            for model_info in self._info.models.values():
                if model_info.schema == model_schema:
                    model = model_info.type
                    data.pop("$schema")
                    break

        # if there is still no model, try to use repository's default model
        if not model and self._info.default_model:
            model = self._info.default_model

        # if we still do not have the model, just use repository's url and let it decide
        if not model:
            create_url = self._info.links.records
            has_metadata = False
        else:
            # take the deposit url from the model
            create_url = self._info.models[model].links.records
            has_metadata = self._info.models[model].metadata

        if has_metadata and "metadata" not in data:
            data = {"metadata": {**data}}
        else:
            data = {**data}

        if community or workflow:
            parent: dict[str, Any] = (
                copy.deepcopy(data.pop("parent")) if "parent" in data else {}
            )
            data["parent"] = parent
            if community:
                assert (
                    "community" not in parent
                ), f"Community already in parent: {parent}"
                parent["communities"] = {"default": community}
            if workflow:
                assert "workflow" not in parent, f"Workflow already in data: {parent}"
                parent["workflow"] = workflow
        data["files"] = {"enabled": files_enabled}

        return await self._connection.post(
            url=create_url,
            json=data,
            idempotent=idempotent,
            result_class=Record,
        )

    @override
    async def read(
        self,
        record_id: RecordId,
        *,
        model: str | None = None,
        status: RecordStatus | None = None,
        query: dict[str, str] | None = None,
    ) -> Record:
        """Read a record from the repository. Please provide either record_id or record_url, not both.

        :param record_id:       the id of the record. Could be either pid or url
        :return:                the record
        """
        record_url = _record_id_to_url(
            self._info, record_id, model or self._model, status or self._status
        )

        if query:
            record_url = record_url.with_query(**query)
        return await self._connection.get(
            url=record_url,
            result_class=Record,
        )

    @override
    async def search(
        self,
        *,
        q: Optional[str] = None,
        page: Optional[int] = None,
        size: Optional[int] = None,
        sort: Optional[str] = None,
        model: str | None = None,
        status: RecordStatus | None = None,
        facets: dict[str, str] | None = None,
    ) -> RecordList:
        """Search for records in the repository."""
        search_url, extra_facets = _get_search_params(self._info, model or self._model, status or self._status)

        query = {**(facets or {}), **extra_facets}
        if q:
            query["q"] = q
        if page is not None:
            query["page"] = str(page)
        if size is not None:
            query["size"] = str(size)
        if sort:
            query["sort"] = sort

        return await self._connection.get(
            url=search_url,
            params=query,
            result_class=RecordList,
        )

    @override
    async def next_page(self, *, record_list: RecordList) -> RecordList:
        """Get the next page of records."""
        next_url = record_list.links.next
        if next_url:
            return await self._connection.get(
                url=next_url,
                result_class=RecordList,
            )
        else:
            return self._empty_record_list()

    def _empty_record_list(self):
        return RecordList(
            links=RESTPaginationLinks(
                self_=None,     # type: ignore
            ),
            hits=RESTHits[Record](total=0, hits=[]),
        )

    @override
    async def previous_page(self, *, record_list: RecordList) -> RecordList:
        """Get the previous page of records."""
        prev_url = record_list.links.prev
        if prev_url:
            return await self._connection.get(
                url=prev_url,
                result_class=RecordList,
            )
        else:
            return self._empty_record_list()

    @override
    @contextlib.asynccontextmanager
    async def scan(  # type: ignore
        self,  #  - as we can not find working mypy type tht would work with parent class
        *,
        q: Optional[str] = None,
        model: str | None = None,
        status: RecordStatus | None = None,
        facets: dict[str, str] | None = None,
    ) -> AsyncGenerator[AsyncIterator[Record], None]:
        """Scan all the records in the repository.

        Tries to return all matched records in a consistent manner
        (using snapshots if they are available), but this behaviour is not guaranteed.

        The implementation might rely on specific sorting, so you should not modify it
        unless you know what you are doing.

        Usage:

        ```
        async with client.scan(...) as records:
            async for record in records:
                print(record)
        ```

        Note: in invenio, we do not have the scan api over the rest, so we simulate it
        by:
            * searching page by page if the number of records < OPENSEARCH_SCAN_WINDOW
            * scanning by date bisection if the number of records > OPENSEARCH_SCAN_WINDOW
        """

        def opensearch_date_serialize(date: datetime) -> str:
            return date.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        def opensearch_date_interval(start: datetime, end: datetime) -> str:
            return f"[{opensearch_date_serialize(start)} TO {opensearch_date_serialize(end)}{'}'}"

        async def paginated_scan(
            q: str | None,
            model: str | None,
            status: RecordStatus | None,
            facets: dict[str, str] | None = None,
        ) -> AsyncIterator[Record]:
            result = await self.search(
                q=q,
                size=OPENSEARCH_SCAN_PAGE,
                model=model,
                status=status,
                facets=facets,
            )
            while result.hits.hits:
                for record in result.hits.hits:
                    yield record
                result = await self.next_page(record_list=result)

        async def scan_via_date_bisect() -> AsyncIterator[Record]:
            result = await self.search(q=q, size=1, facets=dict(sort="oldest"))
            first_created = result.hits.hits[0].created
            result = await self.search(q=q, size=1, facets=dict(sort="newest"))
            last_created = result.hits.hits[0].created

            # make sure that the first_created and last_created are in UTC
            first_created = first_created.astimezone(UTC) - timedelta(seconds=1)
            last_created = last_created.astimezone(UTC) + timedelta(seconds=1)

            # reversed to push the first half at the top
            stack = [
                [date_in_between(first_created, last_created), last_created],
                [first_created, date_in_between(first_created, last_created)],
            ]

            while stack:
                start, end = stack.pop()
                date_query = f"created:{opensearch_date_interval(start, end)}"
                if q:
                    subquery = f"({q}) AND {date_query}"
                else:
                    subquery = date_query
                result = await self.search(
                    q=subquery,
                    size=1,
                    model=model,
                    status=status,
                    facets=facets,
                )

                if result.hits.total <= OPENSEARCH_SCAN_WINDOW:
                    if result.hits.total:
                        async for record in paginated_scan(
                            q=subquery,
                            model=model,
                            status=status,
                            facets=facets,
                        ):
                            yield record
                else:
                    # reversed to push the first half at the top
                    stack.append([date_in_between(start, end), end])
                    stack.append([start, date_in_between(start, end)])

        async def _scan_pages() -> AsyncGenerator[Record, None]:
            # at first check if we can get all records in one paginated listing
            # repository can not handle page_size=0, so we use 1
            result = await self.search(
                q=q,
                size=1,
                model=model,
                status=status,
                facets=facets,
            )
            if result.hits.total <= OPENSEARCH_SCAN_WINDOW:
                if result.hits.total:
                    async for record in paginated_scan(q, model, status, facets):
                        yield record
            else:
                # there is more than OPENSEARCH_SCAN_WINDOW records, that means
                # that we do not know even the result count. We will get to it
                # by scanning the records by date and using bisection.
                async for record in scan_via_date_bisect():
                    yield record

        yield _scan_pages()

    def _etag_headers(self, etag: str | None) -> dict[str, str]:
        """Return the headers with the etag if it was returned by the repository."""
        headers: dict[str, str] = {}
        if etag:
            headers["If-Match"] = etag
        return headers

    @override
    async def update(self, record: Record, *, verify_version: bool = True) -> Record:
        """Update a record in the repository.

        The record must have an id and optionally
        an etag. If the etag is not provided, the record is updated without checking
        the etag. If the etag is provided, the record is updated only if the etag matches
        the current etag of the record in the repository.

        An updated version, as stored in the repository, is returned.

        :param record: record that will be stored to the server
        """
        headers: dict[str, str] = {}
        if verify_version:
            headers.update(self._etag_headers(record.get_etag()))

        ret = await self._connection.put(
            url=record.links.self_,
            json=converter.unstructure(record),  # type: ignore
            headers={
                **headers,
                "Content-Type": "application/json",
            },
            result_class=Record,
        )
        return ret

    @override
    async def delete(
        self,
        record_id_or_record: RecordId | Record,
        *,
        etag: str | None = None,
        status: RecordStatus | None = None,
    ) -> None:
        """Delete a record inside the repository.

        :param record_id: identification of the record. If record_id is passed, you can
                          specify the etag as well
        :param record: record downloaded from the repository
        :param etag: if record_id is specified, delete the record only if the version matches
        """
        if isinstance(record_id_or_record, Record):
            url = record_id_or_record.links.self_
            etag = record_id_or_record._etag    # type: ignore
        else:
            url = _record_id_to_url(self._info, record_id_or_record, self._model, 
                                   status=status or self._status)

        return await self._connection.delete(url=url, headers=self._etag_headers(etag))

    async def publish(self, record: Record) -> Record | Request:
        """Publish a record.

        :param record: record to publish
        """
        return await self._request_op(
            record,
            "publish_draft",
            "publish",
            "post",
            "Can not publish a record",
        )

    async def _request_op(
        self,
        record: Record,
        request_type_id: str,
        link_name: str,
        link_op: str,
        error_msg: str,
    ) -> Record | Request:
        if record.links["applicable-requests"]:
            request_types = await self._requests_client.applicable_requests(record)

            request_type: RequestType | None = next(
                (rt for rt in request_types.hits if rt.type_id == request_type_id), None
            )
            if not request_type:
                raise ValueError(
                    f"{error_msg}: Request type {request_type_id} not found "
                    f"in applicable requests on {record.id}. Run list requests operation "
                    "to get the list of available requests."
                )

            request = await self._requests_client.create(request_type, {}, submit=True)
            if request.status == "accepted":
                if isinstance(request.links.topic, dict):
                    topic_link = request.links.topic["self"]
                else:
                    topic_link = str(request.links.topic)
                return await self.read(topic_link)
            return request
        else:
            return await getattr(self._connection, link_op)(
                url=getattr(record.links, link_name),
                json={},
                result_class=Record,
            )

    async def edit_metadata(self, record: Record) -> Record | Request:
        """Edit metadata of a published record.

        :param record: published record for which metadata will be edited
        """
        return await self._request_op(
            record,
            "edit_published_record",
            "draft",
            "post",
            "Can not edit metadata of a record",
        )

    async def new_version(self, record: Record) -> Record | Request:
        """Edit metadata of a published record.

        :param record: published record for which metadata will be edited
        """
        return await self._request_op(
            record,
            "new_version",
            "versions",
            "post",
            "Can not create a new version of a record",
        )

    async def retract_published(self, record: Record) -> Record | Request:
        """Edit metadata of a published record.

        :param record: published record for which metadata will be edited
        """
        return await self._request_op(
            record,
            "delete_published_record",
            "self_",
            "delete",
            "Can not retract a record",
        )

def date_in_between(date1: datetime, date2: datetime):
    return date1 + (date2 - date1) / 2


def _record_id_to_url(
    info: RepositoryInfo, record_id: RecordId, model: str | None, status: RecordStatus
) -> URL:
    if isinstance(record_id, URL):
        return record_id
    if isinstance(record_id, str) and record_id.startswith("https://"):
        return URL(record_id)
    if model is None and info.default_model:
        model = info.default_model
    if model is None or model == "*":
        base_url = info.links.records
    else:
        base_url = info.models[model].links.records
    ret = base_url / str(record_id)
    if status == RecordStatus.DRAFT:
        ret /= "draft"
    return ret

def _get_search_params(
    info: RepositoryInfo, model: str | None, status: RecordStatus
) -> tuple[URL,dict[str, str]]:
    if model is None and info.default_model:
        model = info.default_model
    if model is None or model == "*":
        if status == RecordStatus.DRAFT:
            draft_url = info.links.drafts
            if draft_url is None:
                raise ValueError("Repository does not support drafts on a generic url")
            return draft_url, {
                "is_published": "false",
            }
        else:
            return info.links.records, {}

    if status == RecordStatus.DRAFT:
        draft_url = info.models[model].links.drafts
        if draft_url is None:
            raise ValueError(f"Model {model} does not support drafts")
        return draft_url, {
            "is_published": "false",
        }
    return info.models[model].links.records, {}



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Commandline client for records."""

from .create import create_record
from .delete import delete_record
from .download import download_record
from .edit_record import edit_record
from .get import get_record
from .publish import publish_record
from .retract import retract_record
from .scan import scan_records
from .search import search_records
from .update import update_record
from .version_record import version_record

__all__ = (
    "download_record",
    "get_record",
    "search_records",
    "create_record",
    "scan_records",
    "update_record",
    "delete_record",
    "publish_record",
    "edit_record",
    "version_record",
    "retract_record",
)



import dataclasses
import enum
import functools
import logging
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any, Optional, overload

import rich_click as click
import yaml
from click.exceptions import Exit

from nrp_cmd.config import Config
from nrp_cmd.errors import RepositoryClientError, RepositoryError, RepositoryJSONError


class VerboseLevel(enum.Enum):
    """Verbosity level."""

    NORMAL = 0
    QUIET = -1
    VERBOSE = 1


class DebugLevel(enum.Enum):
    """Debug level."""

    NONE = 0
    URL = 1
    REQUEST = 2
    RESPONSE = 3


class OutputFormat(enum.Enum):
    """Output format."""

    JSON = "json"
    JSON_LINES = "jsonl"
    YAML = "yaml"
    TABLE = "table"

    def __str__(self):
        """Return the value of the enum."""
        return self.value


@dataclasses.dataclass
class Model:
    draft: bool
    published: bool

    model: Optional[str] = None
    community: Optional[str] = None
    workflow: Optional[str] = None


type ClickCommand = Callable[..., None]


class Argument(click.Argument):
    """A :class:`click.Argument` with help text."""

    def __init__(self, *args: Any, help: str | None = None, **attrs: Any):
        super().__init__(*args, **attrs)
        self.help = help

    def get_argument_help_record(
        self, ctx: click.Context
    ) -> tuple[str, str, str] | None:
        if self.name and self.help:
            return self.name, "" if self.required else "optional", self.help or ""
        return None


argument_with_help = functools.partial(click.argument, cls=Argument)

@dataclasses.dataclass
class Output:
    """Common traits for CLI commands."""

    verbosity: VerboseLevel = VerboseLevel.NORMAL
    progress: bool = False
    output: Optional[Path] = None
    output_format: Optional[OutputFormat] = None


@overload
def with_verbosity(
    arg: VerboseLevel,
) -> Callable[[ClickCommand], ClickCommand]: ...


@overload
def with_verbosity(
    arg: ClickCommand,
) -> ClickCommand: ...


def with_verbosity(
    arg: VerboseLevel | ClickCommand,
) -> Callable[[ClickCommand], ClickCommand] | ClickCommand:
    """Add verbosity and debug options to a command."""
    default_level = arg if isinstance(arg, VerboseLevel) else VerboseLevel.NORMAL

    def decorator(func: ClickCommand) -> ClickCommand:

        @click.option("--verbose", "-v", count=True, help="Increase verbosity")
        @click.option("--quiet", "-q", count=True, help="Decrease verbosity")
        @click.option("--log-url", is_flag=True, help="Log urls")
        @click.option("--log-request", is_flag=True, help="Log requests")
        @click.option("--log-response", is_flag=True, help="Log responses")
        @functools.wraps(func)
        def wrapper(
            verbose: int = 0,
            quiet: int = 0,
            log_url: bool = False,
            log_request: bool = False,
            log_response: bool = False,
            **kwargs: Any,
        ) -> None:
            out: Output = kwargs.pop("out", None) or Output()
            verbose = default_level.value + verbose - quiet
            match verbose:
                case 0:
                    out.verbosity = VerboseLevel.NORMAL
                case _ if verbose > 0:
                    out.verbosity = VerboseLevel.VERBOSE
                case _:
                    out.verbosity = VerboseLevel.QUIET
            if log_url:
                log = logging.getLogger("nrp_cmd.communication.url")
                log.setLevel(logging.INFO)
            if log_request:
                log = logging.getLogger("nrp_cmd.communication.request")
                log.setLevel(logging.INFO)
            if log_response:
                log = logging.getLogger("nrp_cmd.communication.response")
                log.setLevel(logging.INFO)
            func(out=out, **kwargs)

        wrapper.__name__ += "_with_output"
        return wrapper

    if isinstance(arg, VerboseLevel):
        return decorator
    return decorator(arg)


def with_setvar(func: ClickCommand) -> ClickCommand:
    """Add a variable to store the result of the command."""

    @click.option(
        "--set",
        "variable",
        help="Store the result URL(s) in a variable",
    )
    @functools.wraps(func)
    def wrapper(
        variable: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        if variable and variable.startswith("@"):
            variable = variable[1:]
        func(variable=variable, **kwargs)

    wrapper.__name__ += "_with_setvar"
    return wrapper


def with_output(func: ClickCommand) -> ClickCommand:
    """Add output options to a command."""

    @click.option(
        "-o",
        "--output",
        help="Save the output to a file",
        type=click.Path(),
    )
    @click.option(
        "-f",
        "--output-format",
        help="The format of the output",
        type=click.Choice([str(f) for f in OutputFormat]),
    )
    @functools.wraps(func)
    def wrapper(
        output: Optional[Path] = None,
        output_format: Optional[OutputFormat] = None,
        **kwargs: Any,
    ) -> None:
        out = kwargs.pop("out", None) or Output()
        out.output = Path(output) if output else None
        out.output_format = OutputFormat(output_format) if output_format else None
        func(out=out, **kwargs)

    wrapper.__name__ += "_with_output"
    return wrapper


def with_progress(func: ClickCommand) -> ClickCommand:
    """Add a progress bar to a command."""

    @click.option(
        "--progress/--no-progress",
        default=True,
        help="Show progress bar",
    )
    @functools.wraps(func)
    def wrapper(
        progress: bool = True,
        **kwargs: Any,
    ) -> None:
        out = kwargs.pop("out", None) or Output()
        out.progress = progress
        func(out=out, **kwargs)

    wrapper.__name__ += "_with_progress"
    return wrapper


def with_config(func: ClickCommand) -> ClickCommand:
    """Add a config object to a command."""

    @click.option(
        "--config-path",
        help="Path to the configuration file",
        type=click.Path(dir_okay=False, exists=True),
    )
    @functools.wraps(func)
    def wrapper(
        config_path: Optional[Path] = None,
        **kwargs: Any,
    ) -> None:
        _config = Config.from_file(config_path)
        func(config=_config, **kwargs)

    wrapper.__name__ += "_with_config"
    return wrapper


def with_repository(func: ClickCommand) -> ClickCommand:
    @click.option(
        "--repository",
        help="Repository alias",
    )
    @functools.wraps(func)
    def wrapper(
        repository: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        func(repository=repository, **kwargs)

    wrapper.__name__ += "_with_repository"
    return wrapper


def with_resolved_vars(argument_name: str) -> Callable[[ClickCommand], ClickCommand]:
    """Resolve variables in the command arguments."""

    def decorator(func: ClickCommand) -> ClickCommand:

        @functools.wraps(func)
        def wrapper(
            **kwargs: Any,
        ) -> Any:
            if argument_name not in kwargs:
                raise ValueError(f"Argument {argument_name} not found in kwargs {kwargs}")
            config: Config | None = kwargs.get("config")
            if not config:
                raise ValueError(f"Config not found in kwargs {kwargs}")
            variables = config.load_variables()
            resolved: list[str] | str
            if isinstance(kwargs[argument_name], (list, tuple)):
                resolved = []
                for arg in kwargs[argument_name]:
                    if arg.startswith("@"):
                        resolved.extend(variables[arg[1:]])
                    else:
                        resolved.append(arg)
            else:
                arg = kwargs[argument_name]
                if arg.startswith("@"):
                    resolved = config.load_variables()[arg[1:]]
                    if len(resolved) != 1:
                        raise ValueError(f"Variable {arg} has many values: {resolved}")
                    resolved = resolved[0]
                else:
                    resolved = arg

            kwargs[argument_name] = resolved
            func(**kwargs)

        wrapper.__name__ += "_with_resolved_vars"
        return wrapper

    return decorator


@overload
def with_model(
    func: None = None,
    *,
    model: bool = True,
    community: bool = False,
    workflow: bool = False,
    draft: bool = True,
    published: bool = True,
) -> Callable[[ClickCommand], ClickCommand]: ...


@overload
def with_model(
    func: ClickCommand,
    *,
    model: bool = True,
    community: bool = False,
    workflow: bool = False,
    draft: bool = True,
    published: bool = True,
) -> ClickCommand: ...


def with_model(
    func: ClickCommand | None = None,
    *,
    model: bool = True,
    community: bool = False,
    workflow: bool = False,
    draft: bool = True,
    published: bool = True,
) -> Callable[[ClickCommand], ClickCommand] | ClickCommand:
    """Add model options to a command."""

    def decorator(func: ClickCommand) -> ClickCommand:

        @functools.wraps(func)
        def wrapper(
            *,
            model: Optional[str] = None,
            community: Optional[str] = None,
            workflow: Optional[str] = None,
            draft: bool = False,
            published: bool = True,
            **kwargs: Any,
        ) -> None:
            _model = Model(
                model=model,
                community=community,
                workflow=workflow,
                draft=draft,
                published=published,
            )
            func(model=_model, **kwargs)

        if model:
            wrapper = click.option("--model", help="Model name")(wrapper)
        if community:
            wrapper = click.option("--community", help="Community name")(wrapper)
        if workflow:
            wrapper = click.option("--workflow", help="Workflow name")(wrapper)
        if draft:
            wrapper = click.option(
                "--draft/--no-draft", default=False, help="Include only drafts"
            )(wrapper)
        if published:
            wrapper = click.option(
                "--published/--no-published",
                default=False,
                help="Include only published records",
            )(wrapper)

        wrapper.__name__ += "_with_model"
        return wrapper

    if func is None:
        return decorator
    return decorator(func)


def with_record_ids(func: ClickCommand) -> ClickCommand:
    """Add record IDs to a command."""

    @argument_with_help("record_ids", type=str, nargs=-1, help="Record ID(s)")
    @functools.wraps(func)
    def wrapper(
        record_ids: list[str],
        **kwargs: Any,
    ) -> None:
        func(record_ids=record_ids, **kwargs)

    wrapper.__name__ += "_with_record_ids"
    return with_resolved_vars("record_ids")(wrapper)


def _print_error(e: Exception) -> None:
    if isinstance(e, ExceptionGroup):
        for e in e.exceptions:
            _print_error(e)
    elif isinstance(e, RepositoryJSONError):

        if "message" in e.json:
            click.secho(
                f"Client error: {e.json['message']} ({e.json.get('status', 'unknown status')})",
                err=True,
                fg="red",
            )
            click.secho(f"  {e.request_info.url}", err=True, fg="yellow")
        else:
            click.secho(f"Client error: {e.request_info.url}", err=True, fg="red")
        if "errors" in e.json:
            if isinstance(e.json["errors"], list):
                for error in e.json["errors"]:
                    if (
                        isinstance(error, dict)
                        and "field" in error
                        and "messages" in error
                    ):
                        click.secho(f"  {error['field']}:", err=True)
                        click.secho(
                            "    " + "\n    ".join(error["messages"]),
                            err=True,
                            fg="red",
                        )
                    else:
                        yaml.safe_dump(error, sys.stderr)
            else:
                yaml.safe_dump(e.json, sys.stderr)
        else:
            yaml.safe_dump(e.json, sys.stderr)
    elif isinstance(e, RepositoryClientError):
        click.echo(f"Client error: {e}", err=True)
    elif isinstance(e, RepositoryError):
        click.echo(f"Repository error: {e}", err=True)
    else:
        click.echo(f"Error: {e}", err=True)


def with_errors(func: ClickCommand) -> ClickCommand:
    """Add error handling to a command."""

    @click.option(
        "--log-stacktrace", is_flag=True, help="Log stack traces in case of error"
    )
    @functools.wraps(func)
    def wrapper(
        log_stacktrace: bool = False,
        **kwargs: Any,
    ) -> None:
        try:
            func(**kwargs)
        except ExceptionGroup as eg:
            _print_error(eg)
            if log_stacktrace:
                raise
            raise Exit(1) from None
        except Exception as e:
            _print_error(e)
            if log_stacktrace:
                raise
            raise Exit(1) from None

    wrapper.__name__ += "_with_errors"
    return wrapper



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""CLI commands for managing repositories."""

import sys
import urllib
import urllib.parse
from collections.abc import Generator
from functools import partial
from typing import Any

import rich_click as click
from rich import box
from rich.console import Console
from rich.table import Table
from yarl import URL

from nrp_cmd.async_client import AsyncRepositoryClient, get_async_client
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.config import Config
from nrp_cmd.config.repository import RepositoryConfig
from nrp_cmd.converter import converter

from .arguments import (
    Output,
    VerboseLevel,
    argument_with_help,
    with_config,
    with_output,
    with_verbosity,
)
from .base import OutputWriter


@argument_with_help("url", type=str, help="The URL of the repository")
@argument_with_help(
    "alias", type=str, required=False, help="The alias of the repository"
)
@click.option("--token", type=str, help="The token to access the repository")
@click.option(
    "--verify-tls/--no-verify-tls", default=True, help="Verify the TLS certificate"
)
@click.option("--retry-count", type=int, default=5, help="Number of retries")
@click.option(
    "--retry-after-seconds", type=int, default=5, help="Retry after this interval"
)
@click.option(
    "--anonymous/--no-anonymous",
    default=False,
    help="Do not store the token in the configuration",
)
@click.option(
    "--default/--no-default", default=False, help="Set this repository as the default"
)
@click.option(
    "--launch-browser/--no-launch-browser",
    default=True,
    help="Open the browser to create a token",
)
@with_config
@with_verbosity
@async_command
async def add_repository(
    *,
    config: Config,
    out: Output,
    url: str,
    alias: str | None,
    token: str | None,
    verify_tls: bool,
    retry_count: int,
    retry_after_seconds: int,
    anonymous: bool,
    default: bool,
    launch_browser: bool,
) -> None:
    """Add a new repository to the configuration."""
    console = Console()
    console.print()

    if not url.startswith("https://"):
        url = f"https://{url}"

    if alias is None:
        alias = urllib.parse.urlparse(url).netloc

    try:
        config.get_repository(alias)
        console.print(f'[red]Repository with alias "{alias}" already exists[/red]')
        sys.exit(1)
    except KeyError:
        pass

    console.print(
        f"Adding repository [green]{url}[/green] with alias [green]{alias}[/green]"
    )

    if token is None and not anonymous:
        # open  default browser with url to create token

        login_url = f"{url}/account/settings/applications/tokens/new"

        console.print(
            f"\nI will try to open the following url in your browser:\n{login_url}\n",
        )
        console.print(
            "Please log in inside the browser.\nWhen the browser tells you "
            "that the token has been created, \ncopy the token and paste it here.",
        )
        console.print("Press enter to open the page ...")
        click.getchar()

        if launch_browser:
            try:
                click.launch(login_url)
            except Exception:  # noqa
                console.print(
                    "Failed to open the browser. Please open the URL above manually."
                )

        # wait until the token is created at /account/settings/applications/tokens/retrieve
        token = click.prompt("\nPaste the token here").strip()

    console.print("Creating repository with token ...")
    config.add_repository(
        RepositoryConfig(
            alias=alias,
            url=URL(url),
            token=token,
            verify_tls=verify_tls,
            retry_count=retry_count,
            retry_after_seconds=retry_after_seconds,
        )
    )
    if default or len(config.repositories) == 1:
        config.set_default_repository(alias)

    # check if the repository is reachable and get its parameters
    client: AsyncRepositoryClient = await get_async_client(alias, config=config)
    await client.get_repository_info(refresh=True)

    console.print(f"[green]Added repository {alias} -> {url}[/green]")
    config.save()


@argument_with_help("alias", type=str, help="The alias of the repository")
@with_config
@async_command
async def remove_repository(*, config: Config, alias: str) -> None:
    """Remove a repository from the configuration."""
    console = Console()
    console.print()
    config.remove_repository(alias)
    console.print(f'[green]Removed repository "{alias}"[/green]')
    config.save()


@argument_with_help("alias", type=str, help="The alias of the repository")
@with_config
@async_command
async def select_repository(*, config: Config, alias: str) -> None:
    """Select a default repository."""
    console = Console()
    console.print()

    config.set_default_repository(alias)
    console.print(f'[green]Selected repository "{alias}"[/green]')
    config.save()


@argument_with_help("alias", type=str, help="The alias of the repository")
@with_config
@async_command
async def enable_repository(
    *,
    config: Config,
    alias: str,
) -> None:
    """Enable a disabled repository. Enabling it will make it usable for subsequent commands."""
    console = Console()
    console.print()

    for repo in config.repositories:
        if repo.alias == alias:
            repo.enabled = True
            break
    else:
        console.print(f'[red]Repository "{alias}" not found[/red]')
        sys.exit(1)
    console.print(f'[green]Enabled repository "{alias}"[/green]')
    config.save()


@argument_with_help("alias", type=str, help="The alias of the repository")
@with_config
@async_command
async def disable_repository(
    *,
    config: Config,
    alias: str,
) -> None:
    """Disable a repository. This repository will not be used for any operation."""
    console = Console()
    console.print()

    for repo in config.repositories:
        if repo.alias == alias:
            repo.enabled = False
            break
    else:
        console.print(f'[red]Repository "{alias}" not found[/red]')
        sys.exit(1)
    console.print(f'[green]Disabled repository "{alias}"[/green]')
    config.save()


@with_output
@with_verbosity
@with_config
@async_command
async def list_repositories(
    *,
    config: Config,
    out: Output,
) -> None:
    """List all repositories."""
    console = Console()

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(output_tables, verbose=out.verbosity == VerboseLevel.VERBOSE),
    ) as printer:
        printer.output(
            [
                dump_repo_configuration(
                    repo, out.verbosity == VerboseLevel.VERBOSE, config.default_alias
                )
                for repo in config.repositories
            ]
        )


@with_verbosity(VerboseLevel.VERBOSE)
@with_output
@with_config
@argument_with_help("alias", type=str, help="The alias of the repository")
@click.option(
    "--refresh",
    is_flag=True,
    help="Force a refresh of the information",
)
@async_command
async def describe_repository(
    *,
    config: Config,
    out: Output,
    alias: str,
    refresh: bool,
) -> None:
    """Get information about a repository."""
    console = Console()

    client: AsyncRepositoryClient = await get_async_client(alias, config=config)

    repo_config = config.get_repository(alias)

    if refresh:
        repo_config.info = await client.get_repository_info(refresh=refresh)
        config.save()

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        output_repository_info_table,
    ) as printer:
        printer.output(
            dump_repo_configuration(
                repo_config,
                verbose=out.verbosity == VerboseLevel.VERBOSE,
                default_alias=config.default_alias,
            )
        )


def dump_repo_configuration(
    repo: RepositoryConfig, verbose: bool, default_alias: str | None
) -> dict[str, Any]:
    """Dump the repository configuration into a json-compatible structure."""
    if not verbose:
        return {
            "alias": repo.alias,
            "url": str(repo.url),
            "default": repo.alias == default_alias,
        }
    return {
        "alias": repo.alias,
        "url": str(repo.url),
        "token": "***" if repo.token else "anonymous",
        "verify_tls": repo.verify_tls,
        "retry_count": repo.retry_count,
        "retry_after_seconds": repo.retry_after_seconds,
        "info": converter.unstructure(repo.info, keep_nulls=True)
        if repo.info
        else None,
        "default": repo.alias == default_alias,
    }


def output_tables(
    data: Any, *, verbose: bool, **kwargs: Any
) -> Generator[Table, None, None]:
    """Output the information about the repositories formatted as a table."""
    if not verbose:
        table = Table(title="Repositories", box=box.SIMPLE, title_justify="left")
        table.add_column("Alias", style="cyan")
        table.add_column("URL")
        table.add_column("Default", style="green")

        for repo in data:
            table.add_row(
                repo["alias"],
                str(repo["url"]),
                "✓" if repo["default"] else "",
            )
        yield table
    else:
        for repo in data:
            yield from output_repository_info_table(repo, **kwargs)


def output_repository_info_table(
    data: dict[str, Any],
    **kwargs: Any,  # noqa: ANN401
) -> Generator[Table, None, None]:
    """Output the information about a repository formatted as a table."""
    repo = data
    table = Table(
        title=f"Repository '{repo["alias"]}'",
        box=box.SIMPLE,
        show_header=False,
        title_justify="left",
    )
    table.add_column("", style="cyan")
    table.add_column("")
    if repo["info"]:
        table.add_row("Name", repo["info"]["name"])
    table.add_row("URL", str(repo["url"]))
    table.add_row("Token", repo["token"])
    table.add_row("TLS Verify", "✓" if repo["verify_tls"] else "[red]skip[/red]")
    table.add_row("Retry Count", str(repo["retry_count"]))
    table.add_row("Retry After Seconds", str(repo["retry_after_seconds"]))
    table.add_row("Default", "✓" if repo["default"] else "")
    if repo["info"]:
        table.add_row("Version", repo["info"]["version"])
        table.add_row("Invenio Version", repo["info"]["invenio_version"])
        table.add_row("Transfers", ", ".join(sorted(repo["info"]["transfers"])))
        table.add_row("Records url", str(repo["info"]["links"]["records"]))
        table.add_row("User drafts", str(repo["info"]["links"]["drafts"]))
    yield table

    if repo["info"]:
        model_info: Any
        for model_name, model_info in repo["info"]["models"].items():
            table = Table(
                title=f"Model '{model_name}'",
                box=box.SIMPLE,
                show_header=False,
                title_justify="left",
            )
            table.add_column("", style="cyan")
            table.add_column("")
            table.add_row("Name", model_info["name"])
            table.add_row("Description", model_info["description"])
            table.add_row("Version", model_info["version"])
            table.add_row("Features", ", ".join(model_info["features"]))
            table.add_row("HTML", str(model_info["links"]["html"]))
            table.add_row("Model Schema", str(model_info["links"]["model"]))
            table.add_row("Published Records URL", str(model_info["links"]["records"]))
            table.add_row("User Records URL", str(model_info["links"]["drafts"]))

            for ct in sorted(
                model_info["content_types"], key=lambda x: x["content_type"]
            ):
                table.add_row("Content-Type", ct["content_type"])
                table.add_row("", ct["name"] + "\n" + ct["description"])
                table.add_row("    Schema", str(ct["schema"]))
                table.add_row("    Can Export", "✓" if ct["can_export"] else "")
                table.add_row("    Can Deposit", "✓" if ct["can_deposit"] else "")

            yield table



#
# This file was generated from the asynchronous client at base_client.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Sync client for NRP repositories."""
# TODO: can not do from __future__ import annotations here
# as it is not compatible with attrs trying to resolve the
# type annotations in runtime

from collections.abc import Iterator
from contextlib import AbstractContextManager
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Protocol, Self, overload

from yarl import URL

from ..config import RepositoryConfig
from ..types.files import TRANSFER_TYPE_LOCAL, File
from ..types.info import RepositoryInfo
from ..types.records import Record, RecordId, RecordList
from ..types.requests import Request, RequestList, RequestType, RequestTypeList
from .streams import DataSink, DataSource


class RecordStatus(Enum):
    """Selector for records."""

    ALL = "all"
    """All records"""

    PUBLISHED = "published"
    """Published records"""

    DRAFT = "draft"
    """Draft records"""


class SyncRecordsClient(Protocol):
    """Client class for access to records."""

    def with_model(self, model: str) -> Self:
        """Return a new client limited to the given model."""
        ...

    @property
    def published_records(self) -> Self:
        """Return a new client limited to published records."""
        ...

    @property
    def draft_records(self) -> Self:
        """Return a new client limited to draft records."""
        ...

    def create(
        self,
        data: dict[str, Any],
        *,
        model: str | None = None,
        community: str | None = None,
        workflow: str | None = None,
        idempotent: bool = False,
        files_enabled: bool = True,
    ) -> Record:
        """Create a new record in the repository.

        :param data:            the metadata of the record
        :param community:       community in which the record should be created
        :param workflow:        the workflow to use for the record, if not provided
                                the default workflow of the community is used
        :param idempotent:      if True, the operation is idempotent and can be retried on network errors.
                                Use only if you know that the operation is idempotent, for example that
                                you use PID generator that takes the persistent identifier from the data.
        :return:                the created record
        """
        ...

    def read(
        self,
        record_id: RecordId,
        *,
        model: str | None = None,
        status: RecordStatus | None = None,
        query: dict[str, str] | None = None,
    ) -> Record:
        """Read a record from the repository. Please provide either record_id or record_url, not both.

        :param record_id:       the id of the record. Could be either pid or url
        :param model:           optional model of the record
        :param query:           extra arguments to read, repository specific
        :return:                the record
        """
        ...

    def update(self, record: Record, *, verify_version: bool = True) -> Record:
        """Update a record in the repository.

        The record must have an id and optionally
        an etag. If the etag is not provided, the record is updated without checking
        the etag. If the etag is provided, the record is updated only if the etag matches
        the current etag of the record in the repository.

        An updated version, as stored in the repository, is returned.

        :param record: record that will be stored to the server
        :param verify_version: if set to true, verify that the record
                               on the server has not been modified in the meantime
        """
        ...

    def delete(
        self,
        record_id_or_record: RecordId | Record,
        *,
        etag: str | None = None,
        status: RecordStatus | None = None,
    ) -> None:
        """Delete a record inside the repository.

        :param record_id: identification of the record. If record_id is passed, you can
                          specify the etag as well
        :param record: record downloaded from the repository
        :param etag: if record_id is specified, delete the record only if the version matches
        """
        ...

    def search(
        self,
        *,
        q: Optional[str] = None,
        page: Optional[int] = None,
        size: Optional[int] = None,
        sort: Optional[str] = None,
        model: str | None = None,
        status: RecordStatus | None = None,
        facets: dict[str, str] | None = None,
    ) -> RecordList:
        """Search for records in the repository."""
        ...

    def next_page(self, *, record_list: RecordList) -> RecordList:
        """Get the next page of records."""
        ...

    def previous_page(self, *, record_list: RecordList) -> RecordList:
        """Get the previous page of records."""
        ...

    def scan(
        self,
        *,
        q: Optional[str] = None,
        model: str | None = None,
        status: RecordStatus | None = None,
        facets: dict[str, str],
    ) -> AbstractContextManager[Iterator[Record]]:
        """Scan all the records in the repository.

        Tries to return all matched records in a consistent manner
        (using snapshots if they are available), but this behaviour is not guaranteed.

        The implementation might rely on specific sorting, so you should not modify it
        unless you know what you are doing.

        Usage:

        ```
        with client.scan(...) as records:
            for record in records:
                print(record)
        ```

        Note: the with section is required so that we can clean up the resources
        if there is a break in the iteration.
        """
        ...

    def publish(self, record: Record) -> Record | Request:
        """Publish a record.

        :param record: record to publish
        """
        pass

    def edit_metadata(self, record: Record) -> Record | Request:
        """Edit metadata of a published record.

        :param record: published record for which metadata will be edited
        """
        pass

    def new_version(self, record: Record) -> Record | Request:
        """Create a new version of a published record.

        :param record: published record for which the new version will be created
        """
        pass

    def retract_published(self, record: Record) -> Record | Request:
        """Retract a published record.

        :param record: published record which will be retracted
        """
        pass
class SyncFilesClient(Protocol):
    """Client class for accessing files stored with repository records."""

    def list(
        self,
        record_or_url: Record | URL,
    ) -> list[File]:
        """List the files of a record."""
        ...

    @overload
    def read(self, file_url: URL) -> File: ...

    @overload
    def read(self, record: Record, key: str) -> File: ...

    @overload
    def read(self, record_url: URL, key: str) -> File: ...

    def read(  # type: ignore
        self,
        *args: Record | URL | str,
    ):
        """Read file metadata from the repository.

        You can pass either the file url directly, or the record
        (as a retrieved record or its url) and the key of the file.

        :param file_url: url of the file
        :param record: record where the file is stored
        :param record_url: url of the record where the file is stored
        :param key: key of the file
        """
        ...

    def upload(
        self,
        record_or_url: Record | URL,
        key: str,
        metadata: dict[str, Any],
        source: DataSource | str | Path,
        transfer_type: str = TRANSFER_TYPE_LOCAL,
        transfer_metadata: dict[str, Any] | None = None,
        progress: str | None = None,
    ) -> File:
        """Upload a file to the repository.

        :param record_or_url: record or url of the record where the file will be uploaded
        :param key: key of the file
        :param file: file to upload
        :param metadata: metadata of the file
        :param progress: if set, show subprogress with this name
        """
        ...

    @overload
    def download(
        self,
        record: Record,
        key: str,
        sink: DataSink,
        *,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None,
    ) -> None: ...

    @overload
    def download(
        self,
        file_or_url: File | URL,
        sink: DataSink,
        *,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None,
    ) -> None: ...

    def download(  # type: ignore
        self,
        *args: Record | str | DataSink | File | URL,
        parts: int | None = None,
        part_size: int | None = None,
        progress: str | None,
    ) -> None:
        """Download the file to the sink.

        :param file_rec: file to download
        :param sink: sink where to download the file
        :param parts: number of parts to download the file in
        :param part_size: size of the parts
        :param progress: if set, show subprogress with this name
        """
        ...

    def update(self, file: File) -> File:
        """Update the file metadata in the repository.

        :param file: file to update
        """
        ...

    @overload
    def delete(self, record: Record, key: str | None = None) -> None: ...

    @overload
    def delete(self, file: File) -> None: ...

    @overload
    def delete(self, file: URL) -> None: ...

    def delete(  # type: ignore
        self,
        arg: Record | File | URL,
        key: str | None = None,
    ) -> None:
        """Delete a record inside the repository.

        Params:
          - record and key
          - File object
          - URL of the file
        """
        ...

class SyncRequestsClient(Protocol):
    """An abstract client for requests within the NRP repository.

    A request is an operation initiated by the user on a topic.
    It can be a request to publish a record, to delete a record,
    assign a DOI, etc. The request has a recipient (which is determined
    automatically by the server) and a status (which can be created,
    accepted, rejected, cancelled, ...).

    After the request is created, the user can check the status of the request
    and wait until the request is completed. Completion of a request might
    trigger automatic action, like publishing a record or assigning a DOI.
    """

    def applicable_requests(
        self, topic: Record | URL, params: dict[str, str] | None = None
    ) -> RequestTypeList:
        """Return all requests that can be created on a given topic at this moment."""
        ...

    @overload
    def create(
        self,
        type_: RequestType,
        payload: dict[str, Any],
        submit: bool = False,
    ) -> Request: ...

    @overload
    def create(
        self,
        topic: Record | URL,
        type_: str,
        payload: dict[str, Any],
        submit: bool = False,
    ) -> Request: ...

    def create(  # type: ignore
        self,
        *args: Record | URL | RequestType | str | dict[str, Any],
        submit: bool = False,
    ) -> Request:
        """Create a new request of this type."""
        ...

    def all(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Search for all requests the user has access to.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def created(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all requests, that are created but not yet submitted.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def submitted(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all submitted requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def accepted(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all accepted requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def declined(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all declined requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def expired(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all expired requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def cancelled(
        self, *, topic: Record | URL | None = None, params: dict[str, str] | None = None
    ) -> RequestList:
        """Return all cancelled requests.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def read_request(self, request_id: str) -> Request:
        """Read a single request by its id.

        :param topic:  if given, filter the requests by the topic
        :param params: Additional parameters to pass to the search query, see repository docs for possible values
        """
        ...

    def submit(
        self, request: Request | URL, payload: dict[str, Any] | None = None
    ) -> Request:
        """Submit the request.

        The request will be either passed to receivers, or auto-approved
        depending on the current workflow
        """
        ...

    def accept(
        self, request: Request | URL, payload: dict[str, Any] | None = None
    ) -> Request:
        """Accept the submitted request."""
        ...

    def decline(
        self, request: Request | URL, payload: dict[str, Any] | None = None
    ) -> Request:
        """Decline the submitted request."""
        ...

    def cancel(
        self, request: Request | URL, payload: dict[str, Any] | None = None
    ) -> Request:
        """Cancel the request."""
        ...


class SyncRepositoryClient(Protocol):
    """An abstract client for NRP repositories.

    Usually, subclasses of this class are not instantiated directly in your code.
    To get an instance of a repository, use the high-level method `get_sync_client`:

    ```
    my_client = sync_client(config?, url=url, refresh=False/True)
    my_client = sync_client(config?, alias=alias, refresh=False/True)
    ```

    and then use the instance.
    """

    # region info endpoint
    @classmethod
    def can_handle_repository(
        cls, url: URL | str, verify_tls: bool = True
    ) -> URL | None:
        """Return if this client can handle a repository that contains the passed URL.

        This method can make an http request or use any means to check if the repository
        at the URL can be handled by this client.

        :param url: any url within the repository (root, record api, html, documentation
        if running on the same host, ...)
        :param verify_tls: whether to verify tls (should be switched on for production)
        :return: API url of the server or None if this client can not handle the repository
        """
        ...

    @classmethod
    def from_configuration(
        cls,
        config: RepositoryConfig,
        refresh: bool = False,
        extra_tokens: dict[URL, str] | None = None,
    ) -> Self:
        """Create a client from the given configuration.

        :param config: the configuration for the repository
        :param refresh: refresh the configuration by calling get_repository_info
        """
        ...

    def get_repository_info(self, refresh: bool = True) -> RepositoryInfo:
        """Get information about the repository.

        This call is cached inside the RepositoryConfig instance.

        :param refresh: refresh the info from the server
        """
        ...

    # endregion

    @property
    def records(self) -> SyncRecordsClient:
        """Return client for accessing records."""
        ...

    @property
    def files(self) -> SyncFilesClient:
        """Return client for accessing files."""
        ...

    @property
    def requests(self) -> SyncRequestsClient:
        """Return the requests session used by the client."""
        ...




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Different transfers to the invenio repository."""

from .base import Transfer
from .registry import transfer_registry

__all__ = ("Transfer", "transfer_registry")



#
# This file was generated from the asynchronous client at invenio/__init__.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


from .client import SyncInvenioRepositoryClient

__all__ = ("SyncInvenioRepositoryClient", )



"""Types for requests in the NRP repository."""

from datetime import datetime
from enum import StrEnum, auto
from typing import Any, Optional

from attrs import define, field
from cattrs.dispatch import StructureHook
from yarl import URL

from ..converter import Omit, Rename, WrapStructure, extend_serialization
from .base import Model
from .records import BaseRecord
from .rest import RESTList, RESTObject, RESTObjectLinks

# region request types

@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RequestTypeActionLinks(Model):
    """Links on a request type object."""

    create: Optional[URL] = None
    """Link to create a new request of this type."""


@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RequestTypeLinks(Model):
    """Links on a request type."""

    actions: RequestTypeActionLinks = field()
    """Actions that can be performed on the request type."""


@extend_serialization(Omit("_etag", from_unstructure=True), 
                      allow_extra_data=True)
@define(kw_only=True)
class RequestType(RESTObject):
    """A type of request that the user can apply for.

    An example might be a request for access to a dataset,
    publish draft request, assign doi request, ...
    """

    type_id: str = field()
    """Unique identifier of the request type."""

    links: RequestTypeLinks = field()
    """Links on the request type object."""


@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RequestTypeList(RESTList[RequestType]):
    """A list of request types as returned from the API."""

    def __getitem__(self, type_id: str) -> RequestType:
        """Return a request type by its type_id.

        :param type_id:     type_id, stays stable regardless of server version
        :return:            request type or None if not found
        """
        for hit in self.hits:
            if hit.type_id == type_id:
                return hit
        raise KeyError(f"Request type {type_id} not found")

    def __contains__(self, type_id: str) -> bool:
        """Check if a request type with the given type_id is in the list.

        :param type_id:     type_id, stays stable regardless of server version
        :return:            True if the request type is in the list
        """
        return type_id in self.keys()

    def keys(self) -> set[str]:
        """Return all type_ids of the request types in this list.

        :return: a set of type_ids
        """
        return {hit.type_id for hit in self.hits}

    def __getattr__(self, type_id: str) -> RequestType:
        """Return a request type by its type_id.

        Shortcut to be able to write request_types.publish_draft instead of request_types["publish_draft"]
        """
        if type_id in self.keys():
            return self[type_id]
        return super().__getattr__(type_id)

# endregion

# region requests


@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RequestActionLinks(Model):
    """Possible actions on a request."""

    submit: Optional[URL] = None
    cancel: Optional[URL] = None
    accept: Optional[URL] = None
    decline: Optional[URL] = None


@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class RequestLinks(RESTObjectLinks):
    """Links on a request."""

    actions: RequestActionLinks = field()
    """Actions that can be performed on the request at the moment by the current user."""

    comments: URL = field()
    """Link to the comments on the request"""

    timeline: URL = field()
    """Link to the timeline (events) of the request"""


class RequestStatus(StrEnum):
    """Status of the request."""

    CREATED = auto()
    ACCEPTED = auto()
    DECLINED = auto()
    SUBMITTED = auto()
    CANCELLED = auto()
    EXPIRED = auto()


@define(kw_only=True)
class RequestPayloadRecord(Model):
    """A publish/edit/new version request can have a simplified record serialization inside its payload.

    Currently the serialization contains only links to the published/draft record.
    """

    links: RESTObjectLinks
    """Links to the record (self and self_html)"""


def restore_hierarchy(data: dict[str, Any], type_: type, previous: StructureHook) -> Any: # noqa: ANN401
    """Restore the hierarchy of the request payload."""

    def _parse_colon_hierarchy(obj: dict[str, Any], key: str, value: Any) -> None:   # noqa: ANN401
        parts = key.split(":")
        for part in parts[:-1]:
            obj = obj.setdefault(part, {})
        obj[parts[-1]] = value

    if not data:
        return previous(data, type_)
    
    obj: dict[str, Any] = {}
    for k, v in data.items():
        _parse_colon_hierarchy(obj, k, v)
    return previous(obj, type_)

@extend_serialization(WrapStructure(restore_hierarchy), allow_extra_data=True)
@define(kw_only=True)
class RequestPayload(Model):
    """Payload of a request.

    It can be of different types, depending on the request type.
    In the library, the payload is extensible. If you know that there is a specific property
    on the payload, just use payload.property_name to access it.
    """

    published_record: RequestPayloadRecord | None = None
    """A publish request can have a simplified record serialization inside its payload."""

    draft_record: RequestPayloadRecord | None = None
    """An edit request can have a simplified record serialization inside its payload."""


@extend_serialization(allow_extra_data=False)
@define(kw_only=True)
class Request(BaseRecord):
    """Interface for a request in the NRP repository."""

    links: RequestLinks = field()
    """Links on the request object."""

    type: str = field()
    """Request type identifier."""

    title: Optional[str] = None
    """Title of the request, might be None"""

    status: RequestStatus = field()
    """Status of the request"""

    is_closed: bool = field()
    """Is the request closed?"""

    is_open: bool = field()
    """Is the request open?"""

    expires_at: Optional[datetime] = None
    """When the request expires, might be unset"""

    is_expired: bool = field()
    """Is the request expired?"""

    created_by: dict[str, str] = field()
    """Who created the request. It is a dictionary containing a 
    reference to the creator (NOT the links at the moment)."""

    receiver: dict[str, str] = field()
    """Who is the receiver of the request. It is a dictionary containing a 
    reference to the receiver (NOT the links at the moment)."""

    topic: dict[str, str] = field()
    """The topic of the request. It is a dictionary containing a
    reference to the topic (NOT the links at the moment)."""

    payload: Optional[RequestPayload] = None
    """Payload of the request. It can be of different types, depending on the request type."""

    def __attrs_post_init__(self):
        """Check that the created_by, receiver and topic are single values."""
        single_value_expected(self.created_by)
        single_value_expected(self.receiver)
        single_value_expected(self.topic)


@extend_serialization(allow_extra_data=True)
@define(kw_only=True)
class RequestList(RESTList[Request]):
    """A list of requests."""

    sortBy: Optional[str] = None
    """By which property should be the list sorted"""

    aggregations: Optional[Any] = None
    """Aggregations of the list"""



def single_value_expected(value: list[Any] | tuple[Any, ...] | dict[Any, Any]) -> None:
    """Check that the value is a single value."""
    assert len(value) == 1, "Expected exactly one value"



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base rest types."""

# TODO: can not use from __future__ import annotations because of attrs
# currently python is unable to resolve type hints for generic types
# when from __future__ import annotations is used

from collections.abc import Iterator  # noqa TCH003 attrs need to have types in runtime
from datetime import datetime  # noqa TCH003 attrs need to have types in runtime
from types import UnionType
from typing import Any, Optional, Union, get_args, get_origin

from attrs import define, field
from yarl import URL  # noqa TCH003 attrs need to have types in runtime

from ..converter import Rename, converter, extend_serialization, Omit
from .base import Model


@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class RESTObjectLinks(Model):
    """Each rest object must return a links section."""

    self_: URL
    """Link to the object itself (API)"""

    self_html: Optional[URL] = None
    """Link to the object itself (HTML page if it has any)"""


@extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class RESTObject(Model):
    """Base class for all objects returned from the REST API."""

    links: RESTObjectLinks
    """Links to the object itself"""
    
    _etag: Optional[str] = field(default=None, alias='etag', init=False)
    
    def get_etag(self) -> Optional[str]:
        """Return the ETag of the object."""
        return self._etag


@extend_serialization(Rename("self", "self_"), allow_extra_data=True)
@define(kw_only=True)
class RESTPaginationLinks(RESTObjectLinks):
    """Extra links on the pagination response."""

    next: Optional[URL] = None
    """Link to the next page"""

    prev: Optional[URL] = None
    """Link to the previous page"""


@define(kw_only=True)
class RESTHits[T: RESTObject](Model):
    """List of records on the current page."""

    hits: list[T]
    """List of records"""
    
    total: int    


    def __len__(self) -> int:
        """Return the number of records on the current page."""
        return len(self.hits)

    def __iter__(self) -> Iterator[T]:
        """Iterate over the records on the current page."""
        return iter(self.hits)
    
    def __getitem__(self, index: int) -> T:
        """Return the record at the given index."""
        return self.hits[index]

# Note: extending classes need to add this to their decorator
# @extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class RESTList[T: RESTObject](RESTObject):
    """List of REST objects according to the Invenio REST API conventions."""

    links: RESTPaginationLinks = field()
    """Links to the current page, next and previous pages"""

    hits: RESTHits[T] = field(alias='hits')
    """List of records on the current page"""

    @property
    def total(self) -> int:
        """Return the total number of records."""
        return self.hits.total

    def __len__(self) -> int:
        """Return the number of records on the current page."""
        return len(self.hits)

    def __iter__(self) -> Iterator[T]:
        """Iterate over the records on the current page."""
        return iter(self.hits)

    def has_next(self) -> bool:
        """Check if there is a next page."""
        return bool(self.links.next)

    def has_prev(self) -> bool:
        """Check if there is a previous page."""
        return bool(self.links.prev)

type RecordIdType = str | int

@extend_serialization(Omit("_etag", from_unstructure=True), allow_extra_data=True)
@define(kw_only=True)
class BaseRecord(RESTObject):
    """Interface for a record in the NRP repository."""

    id: RecordIdType = field()
    """Identifier of the record"""

    created: datetime = field()
    """Timestamp when the record was created"""

    updated: datetime = field()
    """Timestamp when the record was last updated"""

    revision_id: Optional[int] = None
    """Internal revision identifier of the record"""


def is_record_id(t: Any) -> bool:
    """Return true if given type is record id."""
    origin = get_origin(t)
    if origin not in (UnionType, Union):
        return False
    base_types = get_args(t)
    return set(base_types) == {str, int}


converter.register_structure_hook_func(is_record_id, lambda v, ty: v)
converter.register_unstructure_hook_func(is_record_id, lambda v: v)



#
# This file was generated from the asynchronous client at invenio/transfer/base.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Transfer protocols."""
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from yarl import URL

    from ....progress import ProgressBar
    from ...connection import SyncConnection
    from ..files import File
    from ..source import DataSource


class Transfer(Protocol):
    """Protocol for transferring files to the repository."""

    def prepare(
        self,
        connection: SyncConnection,
        files_link: URL,
        transfer_payload: dict,
        source: DataSource,
    ) -> None:
        """Prepare the transfer.

        :param connection:              connection to the repository
        :param files_link:              link where the files are to be uploaded
        :param transfer_payload:        extra payload for the transfer, might be freely modified by this method
        """
        ...

    def upload(
        self,
        connection: SyncConnection,
        initialized_upload: File,
        source: DataSource,
        progress_bar: ProgressBar,
    ) -> None:
        """Upload the file.

        :param connection:              connection to the repository
        :param initialized_upload:      initialized upload as returned from the repository
        :param file:                    file to be uploaded
        """
        ...

    def get_commit_payload(self, initialized_upload: File) -> dict:
        """Finalize the successful upload.

        There is no method to discard an unsuccessful upload,
        just use the "delete" operation on the failed file.

        :param initialized_upload:      initialized upload as returned from the repository
        :return:                        metadata of the upload as returned from the repository
        """
        ...




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command-line interface for searching records."""

from functools import partial
from typing import Optional

from rich.console import Console

from nrp_cmd.async_client.base_client import RecordStatus
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.base import set_variable as setvar
from nrp_cmd.cli.records.table_formatters import (
    format_record_table,
)
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_setvar,
    with_verbosity,
)
from .search import prepare_records_api


@with_config
@with_repository
@with_output
@with_verbosity
@with_setvar
@with_model(community=True)
@argument_with_help("query", type=str, required=False, help="Query string")
@async_command
async def scan_records(
    *,
    config: Config,
    repository: Optional[str],
    query: Optional[str] = None,
    variable: Optional[str] = None,
    model: Model,
    out: Output,
) -> None:
    """Return all records inside repository that match the given query."""
    console = Console()

    records_api = await prepare_records_api(
        config, model.model, model.draft, model.published, repository
    )

    urls: set[str] = set()

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(format_record_table, verbosity=out.verbosity),  # type: ignore # mypy does not understand this
    ) as printer:
        printer.multiple()

        async with records_api.scan(
            q=query,
            model=model.model,
            status=RecordStatus.PUBLISHED if model.published else RecordStatus.DRAFT,
            facets={},
        ) as scan:
            async for entry in scan:
                link = str(entry.links.self_)
                if link not in urls:
                    printer.output(entry)
                    urls.add(link)

    if variable:
        setvar(config, variable, list(urls))



#
# This file was generated from the asynchronous client at doi.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Resolve DOIs to URLs."""

from yarl import URL

from .connection import SyncConnection


def resolve_doi(connection: SyncConnection, doi: str) -> str:
    """Resolve a DOI to a record URL."""
    data = connection.get(url=URL(f"https://api.datacite.org/dois/{doi}"), result_class=dict)
    return data["data"]["attributes"]["url"]




#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Asynchronous connection for the NRP client using the aiohttp library."""

import asyncio
import contextlib
import inspect
import json as _json
import logging
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
)
from functools import partial
from typing import Any, Literal, Optional, cast, overload

from aiohttp import ClientResponse, ClientSession, TCPConnector
from aiohttp.client_exceptions import ClientConnectorError
from cattrs.dispatch import UnstructureHook
from multidict import CIMultiDictProxy, MultiDictProxy
from yarl import URL

from ...converter import deserialize_rest_response
from ...errors import (
    RepositoryClientError,
    RepositoryCommunicationError,
    RepositoryError,
    RepositoryRetryError,
    StructureError,
    is_instance_of_exceptions,
)
from ...progress import DummyProgressBar, ProgressBar, current_progress
from ..streams.base import DataSink, DataSource
from ..streams.progress import ProgressSink
from .auth import AuthenticatedClientRequest, BearerAuthentication, BearerTokenForHost
from .aws_limits import MINIMAL_DOWNLOAD_PART_SIZE, adjust_download_multipart_params
from .limiter import current_limiter
from .response import RepositoryResponse

log = logging.getLogger("invenio_nrp.async_client.connection")
communication_log_url = logging.getLogger("nrp_cmd.communication.url")
communication_log_request = logging.getLogger("nrp_cmd.communication.request")
communication_log_response = logging.getLogger("nrp_cmd.communication.response")


class try_until_success:

    def __init__(
        self,
        attempts: int,
        retry_interval: int,
        too_many_requests_retry_count: int = 5,
        quiet: bool = False,
    ):
        self.attempts = attempts
        self.attempt = 0
        self.too_many_requests_attempts = too_many_requests_retry_count
        self.done = False
        self.failures: list[Exception] = []
        self.retry_interval = retry_interval
        self.current_sleep_interval = 0
        self.quiet = quiet

    async def __aiter__(self):
        while not self.done and self.attempt < self.attempts:
            yield self
            if not self.done:
                log.error(
                    "Retrying %s of %s. Will retry in %s seconds. Latest error: %s...",
                    self.attempt,
                    self.attempts,
                    self.current_sleep_interval,
                    str(self.failures[-1])[:40],
                )
                await asyncio.sleep(self.current_sleep_interval)

        if self.done:
            return

        if self.failures:
            raise ExceptionGroup("Failures in HTTP transport", self.failures)

    async def __aenter__(self):
        self.attempt += 1

    async def __aexit__(self, _ext, exc, _tb):  # type: ignore
        if exc:
            self.failures.append(cast("Exception", exc))
            if isinstance(exc, RepositoryRetryError):
                # if too many requests, we can repeat,
                # so try again and do not add the attempt
                if self.too_many_requests_attempts:
                    self.too_many_requests_attempts -= 1
                    self.attempt -= 1
                if exc.after_seconds is not None:
                    self.current_sleep_interval = exc.after_seconds + 1
                    return True
                self.current_sleep_interval = self.retry_interval * (self.attempt + 1)
                return True

            if is_instance_of_exceptions(exc, (RepositoryClientError, StructureError)):
                return False
            else:
                self.current_sleep_interval = self.retry_interval * (self.attempt + 1)
                return True

        self.done = True


@contextlib.asynccontextmanager
async def _cast_error() -> AsyncIterator[None]:
    """Catch all errors and cast them to Repository*Error."""
    try:
        yield
    except RepositoryError:
        raise
    except RepositoryRetryError:
        raise
    except ClientConnectorError:
        raise
    except Exception as e:
        raise RepositoryCommunicationError(str(e)) from e


class AsyncConnection:
    """Pre-configured asynchronous http connection."""

    def __init__(
        self,
        *,
        tokens: dict[URL, str] | None = None,
        verify_tls: bool = True,
        retry_count: int = 5,
        retry_after_seconds: int = 1,
    ):
        """Create a new connection with the given configuration."""
        self._verify_tls = verify_tls
        self._retry_count = retry_count
        self._retry_after_seconds = retry_after_seconds

        _tokens: list[BearerTokenForHost] = [
            BearerTokenForHost(host_url=url, token=token)
            for url, token in (tokens or {}).items()
            if token
        ]
        self._auth = BearerAuthentication(_tokens)

    @contextlib.asynccontextmanager
    async def _client(
        self, idempotent: bool = False
    ) -> AsyncGenerator[ClientSession, None]:
        """Create a new session with the repository and configure it with the token.

        :return: A new http client
        """
        """
        Create a new session with the repository and configure it with the token.
        :return: A new http client
        """

        connector = TCPConnector(verify_ssl=self._verify_tls)
        async with ClientSession(
            request_class=AuthenticatedClientRequest,
            response_class=RepositoryResponse,
            connector=connector,
            raise_for_status=False,
        ) as session:
            yield session

    @overload
    async def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: Literal[False],
        **kwargs: Any,  # noqa: ANN401
    ) -> CIMultiDictProxy[str]: ...

    @overload
    async def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: Literal[True],
        **kwargs: Any,  # noqa: ANN401
    ) -> MultiDictProxy[MultiDictProxy[str | URL]]: ...

    async def head(
        self,
        *,
        url: URL,
        use_get: bool = False,
        get_links: bool = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> CIMultiDictProxy[str] | MultiDictProxy[MultiDictProxy[str | URL]]:
        """Perform a HEAD request to the repository.

        :param url:                 the url of the request
        :param idempotent:          True if the request is idempotent, should be for HEAD requests
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    None

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """

        async def _head(
            response: ClientResponse,
        ) -> CIMultiDictProxy[str] | MultiDictProxy[MultiDictProxy[str | URL]]:
            if get_links:
                return response.links
            return response.headers

        with current_progress.short_task():
            if use_get:
                return await self._retried(
                    "GET", url, _head, idempotent=True, headers={"Range": "bytes=0-0"}
                )
            else:
                return await self._retried("HEAD", url, _head, idempotent=True)

    async def get[T](
        self,
        *,
        url: URL,
        result_class: type[T],
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Perform a GET request to the repository.

        :param url:                 the url of the request
        :param idempotent:          True if the request is idempotent, should be for GET requests
        :param result_class:        successful response will be parsed to this class
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network error
        """
        with current_progress.short_task():
            return await self._retried(
                "GET",
                url,
                partial(
                    self._get_call_result,
                    result_class=result_class,
                ),
                idempotent=True,
                **kwargs,
            )

    async def post[T](
        self,
        *,
        url: URL,
        json: dict[str, Any] | list[Any] | None = None,
        data: bytes | None = None,
        idempotent: bool = False,
        result_class: type[T],
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Perform a POST request to the repository.

        :param url:                 the url of the request
        :param json:                the json payload of the request (use exactly one of json or data)
        :param data:                the data payload of the request
        :param idempotent:          True if the request is idempotent, normally should be False
        :param result_class:        successful response will be parsed to this class
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        assert (
            json is not None or data is not None
        ), "Either json or data must be provided"

        with current_progress.short_task():
            return await self._retried(
                "POST",
                url,
                partial(
                    self._get_call_result,
                    result_class=result_class,
                ),
                idempotent=idempotent,
                json=json,
                data=data,
                **kwargs,
            )

    async def put[T](
        self,
        *,
        url: URL,
        json: dict[str, Any] | list[Any] | None = None,
        data: bytes | None = None,
        result_class: type[T],
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        """Perform a PUT request to the repository.

        :param url:                     the url of the request
        :param json:                    the json payload of the request (use exactly one of json or data)
        :param data:                    the data payload of the request
        :param result_class:            successful response will be parsed to this class
        :param kwargs:                  any kwargs to pass to the aiohttp client
        :return:                        the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        assert (
            json is not None or data is not None
        ), "Either json or data must be provided"

        with current_progress.short_task():
            return await self._retried(
                "PUT",
                url,
                partial(
                    self._get_call_result,
                    result_class=result_class,
                ),
                idempotent=True,
                json=json,
                data=data,
                **kwargs,
            )

    async def put_stream(
        self,
        *,
        url: URL,
        source: DataSource,
        open_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> ClientResponse:
        """Perform a PUT request to the repository with a file.

        :param url:                 the url of the request
        :param file:                the file to send
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    the response (not parsed)

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """

        async def _put(response: ClientResponse) -> ClientResponse:
            return response

        with current_progress.short_task():
            return await self._retried(
                "PUT",
                url,
                _put,
                idempotent=True,
                data=partial(source.open, **(open_kwargs or {})),
                **kwargs,
            )

    async def get_stream(
        self,
        *,
        url: URL,
        sink: DataSink,
        offset: int = 0,
        size: int | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Perform a GET request to the repository and write the response to a sink.

        :param url:                 the url of the request
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network error
        """

        async def _copy_stream(response: ClientResponse) -> None:
            chunk = await sink.open_chunk(offset=offset)
            try:
                async for data in response.content.iter_any():
                    await chunk.write(data)
            finally:
                await chunk.close()

        if size is not None:
            range_header = f"bytes={offset}-{offset + size - 1}"
        else:
            range_header = f"bytes={offset}-"

        with current_progress.short_task():
            await self._retried(
                "GET",
                url,
                _copy_stream,
                idempotent=True,
                headers={"Range": range_header},
                **kwargs,
            )

    async def delete(
        self,
        *,
        url: URL,
        idempotent: bool = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Perform a DELETE request to the repository.

        :param url:                 the url of the request
        :param idempotent:          True if the request is idempotent, normally should be False
        :param kwargs:              any kwargs to pass to the aiohttp client
        :return:                    None

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        with current_progress.short_task():
            return await self._retried(
                "DELETE", url, None, idempotent=idempotent, **kwargs
            )

    @overload
    async def _get_call_result[T](
        self,
        response: ClientResponse,
        result_class: type[T],
    ) -> T: ...

    @overload
    async def _get_call_result(
        self,
        response: ClientResponse,
        result_class: None,
    ) -> None: ...

    async def _get_call_result[T](
        self,
        response: ClientResponse,
        result_class: type[T] | None,
    ) -> T | None:
        """Get the result from the response.

        :param response:            the aiohttp response
        :param result_class:        the class to parse the response to
        :return:                    the parsed result

        :raises RepositoryClientError: if the request fails due to client passing incorrect parameters (HTTP 4xx)
        :raises RepositoryServerError: if the request fails due to server error (HTTP 5xx)
        :raises RepositoryCommunicationError: if the request fails due to network
        """
        if response.status != 204:
            json_payload = await response.read()
        else:
            json_payload = b""

        if communication_log_response.isEnabledFor(logging.INFO):
            communication_log_response.info(
                "%s", _json.dumps(_json.loads(json_payload))
            )

        await response.raise_for_invenio_status()  # type: ignore
        if response.status == 204:
            assert result_class is None
            return None

        assert result_class is not None
        if issubclass(result_class, ClientResponse):
            return cast("T", response)  # mypy can not get it
        elif issubclass(result_class, str):
            return cast("T", json_payload.decode("utf-8"))  # mypy can not get it
        elif issubclass(result_class, dict):
            return _json.loads(json_payload)
        etag = remove_quotes(response.headers.get("ETag"))
        return deserialize_rest_response(self, json_payload, result_class, etag)

    @overload
    async def _retried[T](
        self,
        method: str,
        url: URL,
        callback: Callable[[ClientResponse], Awaitable[T]],
        idempotent: bool,
        **kwargs: Any,  # noqa: ANN401
    ) -> T: ...

    @overload
    async def _retried(
        self,
        method: str,
        url: URL,
        callback: None,
        idempotent: bool,
        **kwargs: Any,  # noqa: ANN401
    ) -> None: ...

    async def _retried[T](
        self,
        method: str,
        url: URL,
        callback: Callable[[ClientResponse], Awaitable[T]] | None,
        idempotent: bool,
        **kwargs: Any,  # noqa: ANN401
    ) -> T | None:
        """Log the start of a request and retry it if necessary."""
        json = kwargs.get("json")
        if json is not None and callable(json):
            json = json()
            kwargs["json"] = json

        data = kwargs.get("data")

        async for attempt in try_until_success(
            self._retry_count if idempotent else 1, self._retry_after_seconds
        ):
            actual_data = None
            if (
                data is not None
                and callable(data)
                and inspect.iscoroutinefunction(data)
            ):
                actual_data = await data()
                kwargs["data"] = actual_data
            try:

                @contextlib.asynccontextmanager
                async def print_log():
                    if communication_log_url.isEnabledFor(logging.INFO):
                        communication_log_url.info("%s %s", method.upper(), url)
                    if communication_log_request.isEnabledFor(logging.INFO):
                        if json is not None:
                            communication_log_request.info("%s", _json.dumps(json))
                        if data is not None:
                            communication_log_request.info("(stream)")
                    yield

                async with (
                    attempt,
                    current_limiter.limit(url),
                    self._client(idempotent=True) as client,
                    _cast_error(),
                    print_log(),
                    client.request(method, url, auth=self._auth, **kwargs) as response,
                ):
                    if callback is not None:
                        return await callback(response)
                    else:
                        await response.raise_for_invenio_status()  # type: ignore
                    return None
            finally:
                if actual_data is not None and hasattr(actual_data, "close"):
                    await actual_data.close()

        raise Exception("unreachable")

    async def download_file(
        self,
        url: URL,
        sink: DataSink,
        parts: int | None = None,
        part_size: int | None = None,
        progress_bar: ProgressBar | None = None,
    ) -> None:
        progress_bar = progress_bar or DummyProgressBar()

        try:
            headers = await self.head(url=url, get_links=False)
        except RepositoryClientError:
            # The file is not available for HEAD. This is the case for S3 files
            # where the file is a pre-signed request. We'll try to download the headers
            # with a GET request with a range header containing only the first byte.
            headers = await self.head(url=url, use_get=True, get_links=False)

        size = 0
        location = URL(headers.get('Location', url))

        if "Content-Range" in headers:
            size = int(headers["Content-Range"].split("/")[-1])
        elif "Content-Length" in headers:
            size = int(headers["Content-Length"])

        if size:
            await sink.allocate(size)
            progress_bar.set_total(size)

        if (
            size
            and size > MINIMAL_DOWNLOAD_PART_SIZE
            and any(x == "bytes" for x in headers.getall("Accept-Ranges", []))
        ):
            await self._download_multipart(
                location, sink, size, progress_bar, parts, part_size
            )
        else:
            await self._download_single(location, sink, progress_bar)

    async def _download_single(
        self, url: URL, sink: DataSink, progress_bar: ProgressBar
    ) -> None:
        await self.get_stream(url=url, sink=ProgressSink(sink, progress_bar), offset=0)

    async def _download_multipart(
        self,
        url: URL,
        sink: DataSink,
        size: int,
        progress_bar: ProgressBar,
        parts: int | None = None,
        part_size: int | None = None,
    ) -> None:
        adjusted_part_size, adjusted_parts = adjust_download_multipart_params(
            size, parts, part_size
        )

        async with asyncio.TaskGroup() as tg:
            for i in range(adjusted_parts):
                start = i * adjusted_part_size
                part_size = min((i + 1) * adjusted_part_size, size) - start
                tg.create_task(
                    self.get_stream(
                        url=url,
                        sink=ProgressSink(sink, progress_bar),
                        offset=start,
                        size=part_size,
                    )
                )


def remove_quotes(etag: str | None) -> Optional[str]:
    """Remove quotes from an etag.

    :param etag:    the etag header
    :return:        the etag without quotes
    """
    if etag is None:
        return None
    if etag.startswith("W/"):
        etag = etag[2:]
    return etag.strip('"')


def connection_unstructure_hook(data: Any, previous: UnstructureHook) -> Any:
    ret = previous(data)
    ret.pop("_connection", None)
    ret.pop("_etag", None)
    return ret


__all__ = ("AsyncConnection",)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Data sources and sinks."""

from .base import DataSink, DataSource, InputStream, OutputStream, SinkState
from .file import FileSink, FileSource
from .memory import MemorySink, MemorySource
from .stdin import StdInDataSource

__all__ = (
    "DataSink",
    "DataSource",
    "SinkState",
    "InputStream",
    "OutputStream",
    "MemorySink",
    "MemorySource",
    "FileSink",
    "FileSource",
    "StdInDataSource",
)


#
# This file was generated from the asynchronous client at streams/__init__.py by generate_synchronous_client.sh
# Do not edit this file directly, instead edit the original file and regenerate this file.
#


"""Data sources and sinks."""

from .base import DataSink, DataSource, InputStream, OutputStream, SinkState
from .file import FileSink, FileSource
from .memory import MemorySink, MemorySource
from .stdin import StdInDataSource

__all__ = (
    "DataSink",
    "DataSource",
    "SinkState",
    "InputStream",
    "OutputStream",
    "MemorySink",
    "MemorySource",
    "FileSink",
    "FileSource",
    "StdInDataSource",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Authentication for the synchronous client."""

import requests.auth
from yarl import URL

from ...types.auth import BearerTokenForHost


class BearerAuthentication(requests.auth.AuthBase):
    """Bearer token authentication for requests."""

    def __init__(self, tokens: list[BearerTokenForHost]):
        """Initialize the authentication with the tokens."""
        self.tokens = tokens

    def __call__(self, r: requests.Request) -> requests.Request:
        """Add the Authorization header to the request."""
        url = URL(r.url)

        for token in self.tokens:
            if url.host == token.host_url.host and url.scheme == token.host_url.scheme:
                r.headers["Authorization"] = f"Bearer {token.token}"
                break
        return r



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Table formatters for records."""

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any

from rich import box
from rich.table import Table

from nrp_cmd.cli.base import write_table_row
from nrp_cmd.converter import converter

from ..arguments import VerboseLevel

if TYPE_CHECKING:
    from nrp_cmd.types.records import Record, RecordList


def format_search_table(
    data: RecordList, *, verbosity: VerboseLevel
) -> Generator[Table | str, None, None]:
    """Format a search result as a table."""
    table = Table(
        title="Records", box=box.SIMPLE, title_justify="left", show_header=False
    )
    table.add_row("Self", str(data.links.self_))
    table.add_row("Next", str(data.links.next))
    table.add_row("Previous", str(data.links.prev))
    table.add_row("Total", str(data.total))
    yield table

    for record in data:
        yield from format_record_table(record, verbosity=verbosity)


def format_record_table(
    data: Record,
    *,
    verbosity: VerboseLevel,
    **kwargs: Any,  # noqa: ANN401
) -> Generator[Table | str, None, None]:
    """Format a record as a table."""
    if verbosity == VerboseLevel.QUIET:
        yield str(data.links.self_)
    else:
        table = Table(f"Record {data.id}", box=box.SIMPLE, title_justify="left")
        record_dump = converter.unstructure(data)

        if verbosity == VerboseLevel.VERBOSE or "metadata" not in record_dump:
            for k, v in record_dump.items():
                if k != "metadata":
                    write_table_row(table, k, v)

        if "metadata" in record_dump:
            if verbosity == VerboseLevel.VERBOSE:
                table.add_row("Metadata", "")
                prefix = "    "
            else:
                prefix = ""
            for k, v in record_dump["metadata"].items():
                write_table_row(table, k, v, prefix=prefix)

        if verbosity != VerboseLevel.VERBOSE and "errors" in record_dump:
            error_table = Table(box=None, show_header=False)
            for err in record_dump["errors"]:
                error_table.add_row(
                    err["field"], "\n".join(str(x) for x in err["messages"])
                )
            table.add_row("Errors", error_table)

        yield table



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""CLI for requests."""

from .accept import accept_request
from .cancel import cancel_request
from .create import create_request
from .decline import decline_request
from .list import list_requests
from .submit import submit_request

__all__ = (
    "list_requests",
    "create_request",
    "accept_request",
    "decline_request",
    "cancel_request",
    "submit_request",
)



"""Helper functions for requests"""

from yarl import URL

from nrp_cmd.async_client import (
    AsyncRequestsClient,
    get_async_client,
    get_repository_from_record_id,
)
from nrp_cmd.async_client.connection import AsyncConnection
from nrp_cmd.config import Config


async def resolve_request(
    request_id: str, config: Config, repository: str | None = None
) -> tuple[AsyncRequestsClient, URL]:

    connection = AsyncConnection()

    final_request_id, repository_config = await get_repository_from_record_id(
        connection, request_id, config, repository
    )

    config.add_repository(repository_config)
    repository_client = await get_async_client(repository, config=config)

    request_url: URL
    if isinstance(final_request_id, str):
        if final_request_id.startswith("https://"):
            request_url = URL(final_request_id)
        else:
            request_url = repository_config.requests_url / final_request_id
    else:
        request_url = final_request_id

    return repository_client.requests, request_url



import math

# Constants
MAX_UPLOAD_OBJECT_SIZE = 5 * 1024**4           # 5 TiB
MAX_UPLOAD_PARTS = 10_000
MIN_UPLOAD_PART_SIZE = 5 * 1024 * 1024         # 5 MiB
MAX_UPLOAD_PART_SIZE = 5 * 1024**3            # 5 GiB

MINIMAL_DOWNLOAD_PART_SIZE = 1024 * 1024  # 1 MiB
MAXIMAL_DOWNLOAD_PARTS = 10_000


def adjust_upload_multipart_params(size: int, parts: int | None =None, part_size: int | None=None) -> tuple[int, int]:
    """Get multipart params consistent with AWS multipart upload constraints.
    
    Given the total 'size' of an S3 object (in bytes), and optionally
    the desired 'parts' (number of parts) and/or 'part_size' (bytes),
    return a tuple (parts, part_size) that satisfy AWS multipart
    upload constraints.

    AWS Constraints:
    - Max object size: 5 TiB
    - Max number of parts: 10,000
    - Part sizes between 5 MiB and 5 GiB (except the last part can be smaller)
    """
    # 1. Validate the total size
    if size < 0:
        raise ValueError("Size cannot be negative.")
    if size == 0:
        # Edge case: a zero-byte object can be uploaded in a single part
        return (1, 0)  # 0-byte part is valid in practice for a zero-byte file

    if size > MAX_UPLOAD_OBJECT_SIZE:
        raise ValueError(f"Size exceeds maximum S3 object size of {MAX_UPLOAD_OBJECT_SIZE} bytes (5 TiB).")

    # Helper to clamp a part_size within valid range
    def clamp_part_size(ps: int) -> int:
        return max(MIN_UPLOAD_PART_SIZE, min(ps, MAX_UPLOAD_PART_SIZE))

    # We will compute final_parts and final_part_size in a systematic way.
    # Start with any user-provided values, but adjust them to satisfy constraints.

    # Case A: If both parts and part_size are provided
    if parts is not None and part_size is not None:
        # Ensure parts is within [1, 10_000]
        if parts < 1 or parts > MAX_UPLOAD_PARTS:
            raise ValueError(f"'parts' must be in the range [1, {MAX_UPLOAD_PARTS}].")

        # Clamp part_size into [5 MiB, 5 GiB]
        part_size = clamp_part_size(part_size)

        # Check how many parts would actually be needed if we use this part_size
        needed_parts = math.ceil(size / part_size)

        if needed_parts > MAX_UPLOAD_PARTS:
            raise ValueError(
                f"With part_size={part_size} bytes, you would need {needed_parts} parts, "
                f"which exceeds the max of {MAX_UPLOAD_PARTS}."
            )

        if needed_parts != parts:
            parts = needed_parts

        return (parts, part_size)

    # Case B: If only 'parts' is provided
    elif parts is not None:
        if parts < 1 or parts > MAX_UPLOAD_PARTS:
            raise ValueError(f"'parts' must be in the range [1, {MAX_UPLOAD_PARTS}].")

        # Compute the part_size needed for that many parts
        rough_part_size = math.ceil(size / parts)

        # Clamp it to valid range
        part_size_clamped = clamp_part_size(rough_part_size)

        # Recompute the number of parts with the clamped size
        final_parts = math.ceil(size / part_size_clamped)

        if final_parts > MAX_UPLOAD_PARTS:
            raise ValueError(
                f"Requested {parts} parts but cannot satisfy with AWS limits. "
                "Try fewer parts or do not specify 'parts'."
            )

        return (final_parts, part_size_clamped)

    # Case C: If only 'part_size' is provided
    elif part_size is not None:
        # Clamp the requested part_size to [5 MiB, 5 GiB]
        part_size_clamped = clamp_part_size(part_size)

        # Calculate how many parts that implies
        final_parts = math.ceil(size / part_size_clamped)

        if final_parts > MAX_UPLOAD_PARTS:
            # We need to increase part_size to reduce the number of parts
            # part_size_needed >= size / MAX_PARTS
            new_part_size = math.ceil(size / MAX_UPLOAD_PARTS)

            # Clamp again
            new_part_size = clamp_part_size(new_part_size)
            final_parts = math.ceil(size / new_part_size)

            if final_parts > MAX_UPLOAD_PARTS:
                raise ValueError(
                    "Cannot upload within AWS constraints even after adjusting part_size. "
                    f"size={size}, part_size_requested={part_size}, new_part_size={new_part_size}, parts={final_parts}"
                )

            return (final_parts, new_part_size)
        else:
            return (final_parts, part_size_clamped)

    # Case D: Neither 'parts' nor 'part_size' provided
    else:
        # Strategy: pick the maximum possible number of parts up to 10,000
        # i.e., start with the smallest valid part size (5 MiB),
        # and only increase it if we exceed 10,000 parts.

        max_parts_possible = math.ceil(size / MIN_UPLOAD_PART_SIZE)

        if max_parts_possible <= MAX_UPLOAD_PARTS:
            # We can use the minimum part size of 5 MiB and still not exceed 10,000 parts
            final_parts = max_parts_possible
            chosen_part_size = MIN_UPLOAD_PART_SIZE
        else:
            # 5 MiB parts would create too many parts (>10,000).
            # We must increase part_size enough to reduce the count to 10,000 or fewer.
            chosen_part_size = math.ceil(size / MAX_UPLOAD_PARTS)
            chosen_part_size = clamp_part_size(chosen_part_size)  # still enforce [5MiB..5GiB]

            final_parts = math.ceil(size / chosen_part_size)
            if final_parts > MAX_UPLOAD_PARTS:
                raise ValueError(
                    "Cannot satisfy AWS multipart constraints with 10,000 parts. "
                    f"Size={size}, chosen_part_size={chosen_part_size}, parts={final_parts}"
                )

        return (final_parts, chosen_part_size)


def adjust_download_multipart_params(
    size: int,
    parts: int | None = None,
    part_size: int | None = None
) -> tuple[int, int]:
    """Adjust download multipart parameters to fit within constraints.
    
    Computes and returns (final_part_size, final_parts) such that:
      - final_part_size >= MINIMAL_PART_SIZE
      - final_parts <= MAXIMAL_PARTS
      - size fits into final_parts * final_part_size
      - Prefers more parts (smaller part_size) when possible.
    
    :param size: Total size we want to split.
    :param maximal_parts: The largest allowed number of parts.
    :param parts: Desired (or initial) number of parts, can be None.
    :param part_size: Desired (or initial) size of each part, can be None.

    :return (final_part_size, final_parts)

    """

    # Helper that clamps final_part_size to respect minimal size and ensures final_parts <= MAXIMAL_PARTS
    def compute_with_part_size(psize: int) -> tuple[int, int]:
        """Enforce constraints on part_size and compute parts needed."""
        # Enforce minimal part size
        psize = max(psize, MINIMAL_DOWNLOAD_PART_SIZE)
        # Compute parts needed
        needed_parts = math.ceil(size / psize)
        # If we exceed the max, bump the part size up to reduce parts
        if needed_parts > MAXIMAL_DOWNLOAD_PARTS:
            psize = math.ceil(size / MAXIMAL_DOWNLOAD_PARTS)
            needed_parts = math.ceil(size / psize)
        return psize, needed_parts

    # CASE 1: If BOTH part_size and parts are None => 
    #         start from the minimal part size (=> maximum number of parts),
    #         and adjust if that exceeds MAXIMAL_PARTS.
    if part_size is None and parts is None:
        return compute_with_part_size(MINIMAL_DOWNLOAD_PART_SIZE)

    # CASE 2: If part_size is given but parts is None => 
    #         rely on that part_size, adjust if we exceed constraints.
    if part_size is not None and parts is None:
        return compute_with_part_size(part_size)

    # CASE 3: If part_size is None but parts is given => 
    #         we pick the part_size that roughly matches the desired parts, 
    #         then adjust to constraints.
    if part_size is None and parts is not None:
        # The part size that (ideally) yields exactly 'parts'
        candidate_size = math.floor(size / parts)
        if candidate_size < 1:
            # edge case: if parts > size, candidate_size can be 0 => clamp to 1 or minimal
            candidate_size = 1
        return compute_with_part_size(candidate_size)

    # CASE 4: If BOTH are given => the question doesn't define a perfect tie-break. 
    #         We'll treat 'part_size' as the stronger hint (since it directly 
    #         influences how big each chunk is). Then we adjust if needed.
    #         'parts' is effectively ignored except in commentary or debugging.
    return compute_with_part_size(part_size or MINIMAL_DOWNLOAD_PART_SIZE)


# -------------- TEST EXAMPLES ----------------
if __name__ == "__main__":
    # Example usage
    total_size = 50 * 1024 * 1024  # 50 MB

    # No parts or part_size specified
    p, ps = adjust_upload_multipart_params(total_size)
    print("No hints:", p, ps)

    # Only parts specified
    p, ps = adjust_upload_multipart_params(total_size, parts=2)
    print("Only parts=2:", p, ps)

    # Only part_size specified
    p, ps = adjust_upload_multipart_params(total_size, part_size=5 * 1024 * 1024)  # 5 MB
    print("Only part_size=5MB:", p, ps)

    # Both parts and part_size specified
    p, ps = adjust_upload_multipart_params(total_size, parts=3, part_size=6 * 1024 * 1024)
    print("parts=3, part_size=6MB:", p, ps)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command-line interface for searching records."""

from functools import partial
from typing import Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client import AsyncRecordsClient, RecordStatus, get_async_client
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.base import set_variable as setvar
from nrp_cmd.config import Config

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_setvar,
    with_verbosity,
)
from .table_formatters import (
    format_search_table,
)


@argument_with_help("query", type=str, required=False, help="Query string")
@click.option(
    "--size", type=int, default=10, help="Number of results to return on a page"
)
@click.option("--page", type=int, default=1, help="Page number")
@click.option("--sort", type=str, default="bestmatch", help="Sort order")
@with_config
@with_repository
@with_output
@with_verbosity
@with_setvar
@with_model(community=True)
@async_command
async def search_records(
    *,
    config: Config,
    repository: Optional[str] = None,
    query: Optional[str] = None,
    variable: Optional[str] = None,
    model: Model,
    size: int = 10,
    page: int = 1,
    sort: str = "bestmatch",
    out: Output,
) -> None:
    """Return a page of records inside repository that match the given query."""
    console = Console()

    records_api = await prepare_records_api(
        config, model.model, model.draft, model.published, repository
    )

    record_list = await records_api.search(
        q=query,
        page=page,
        size=size,
        sort=sort,
        model=model.model,
        status=RecordStatus.PUBLISHED if model.published else RecordStatus.DRAFT,
        facets={},
    )

    if variable:
        urls = [str(record.links.self_) for record in record_list]
        setvar(config, variable, urls)

    with OutputWriter(
        out.output,
        out.output_format,
        console,
        partial(format_search_table, verbosity=out.verbosity),  # type: ignore # mypy does not understand this
    ) as printer:
        del record_list.aggregations
        printer.output(record_list)


async def prepare_records_api(
    config: Config,
    model: str | None,
    draft: bool,
    published: bool,
    repository: str | None,
) -> AsyncRecordsClient:
    """Get records API for the model and publish status."""
    client = await get_async_client(repository, config=config)
    records_api: AsyncRecordsClient = client.records
    if model is not None:
        records_api = records_api.with_model(model)
    # TODO: published AND draft should be allowed and use the new /all/ endpoint
    if published or not draft:
        records_api = records_api.published_records
    if draft:
        records_api = records_api.draft_records
    return records_api



import math
import re

from yarl import URL

from .types.info import (
    ModelInfo,
    ModelInfoContentType,
    ModelInfoLinks,
    RepositoryInfo,
    RepositoryInfoLinks,
)


def _generate_rdm_vocabulary(
    url: URL, vocabulary_type: str, vocabulary_name: str, special: bool, schema: str
) -> ModelInfo:
    url_prefix = url / "api" if special else url / "api" / "vocabularies"
    return ModelInfo(
        schema=schema,
        type=vocabulary_type,
        name=vocabulary_name,
        description="Vocabulary for " + vocabulary_name,
        version="unknown",
        features=[
            "rdm",
        ],
        links=ModelInfoLinks(
            records=url_prefix / vocabulary_type,
        ),
        content_types=[
            ModelInfoContentType(
                content_type="application/json",
                name="Invenio RDM JSON",
                description="Vocabulary JSON",
                schema=None,
                can_export=True,
                can_deposit=False,
            )
        ],
        metadata=False
    )


def make_rdm_info(url: URL, verify_tls: bool = True) -> RepositoryInfo:
    """If repository does not provide the info endpoint, we assume it is a plain invenio rdm."""
    url = url.with_path("/")
    import requests

    homepage = requests.get(str(url), verify=verify_tls).text
    grp = re.search('meta name="generator" content="InvenioRDM ([0-9.]+)"', homepage)
    if grp:
        rdm_version = math.floor(float(grp.group(1)))
    else:
        raise ValueError("Could not determine Invenio RDM version from homepage")

    transfers = ["L"]
    if rdm_version >= 13:
        transfers.extend(["F", "R", "M"])

    return RepositoryInfo(
        schema="local://introspection-v1.0.0.json",
        name="RDM repository",
        description="",
        version="RDM",
        invenio_version=f"RDM {rdm_version}",
        transfers=transfers,
        links=RepositoryInfoLinks(
            self_=url,
            records=url / "api" / "records",
            drafts=url / "api" / "user" / "records",
            requests=url / "api" / "requests",
            models=None,
        ),
        default_model="records",
        models={
            "records": ModelInfo(
                schema="local://records-v6.0.0.json",
                type="rdm-records",
                name="RDM Records",
                description="RDM Records",
                version="unknown",
                features=[
                    "rdm",
                ],
                links=ModelInfoLinks(
                    html=url / "records",
                    records=url / "api" / "records",
                    drafts=url / "api" / "user" / "records",
                    model=None,
                ),
                content_types=[
                    ModelInfoContentType(
                        content_type="application/json",
                        name="Invenio RDM JSON",
                        description="Invenio RDM JSON as described in "
                        "https://inveniordm.docs.cern.ch/reference/metadata/",
                        schema=url / "schemas" / "records" / "record-v6.0.0.json",
                        can_export=True,
                        can_deposit=True,
                    )
                ],
                metadata=True,
            ),
            "affiliations": _generate_rdm_vocabulary(
                url,
                "affiliations",
                "Affiliations",
                special=True,
                schema="local://affiliation-v1.0.0.json",
            ),
            "awards": _generate_rdm_vocabulary(
                url,
                "awards",
                "Awards",
                special=True,
                schema="local://award-v1.0.0.json",
            ),
            "funders": _generate_rdm_vocabulary(
                url,
                "funders",
                "Funders",
                special=True,
                schema="local://funder-v1.0.0.json",
            ),
            "subjects": _generate_rdm_vocabulary(
                url,
                "subjects",
                "Subjects",
                special=True,
                schema="local://subject-v1.0.0.json",
            ),
            "names": _generate_rdm_vocabulary(
                url,
                "names",
                "Names",
                special=True,
                schema="local://name-v1.0.0.json",
            ),
            "licenses": _generate_rdm_vocabulary(
                url,
                "licenses",
                "Licenses",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "resource_types": _generate_rdm_vocabulary(
                url,
                "resource_types",
                "Resource Types",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "languages": _generate_rdm_vocabulary(
                url,
                "languages",
                "Languages",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "titletypes": _generate_rdm_vocabulary(
                url,
                "titletypes",
                "Title Types",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "creatorsroles": _generate_rdm_vocabulary(
                url,
                "creatorsroles",
                "Creators Roles",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "contributorsroles": _generate_rdm_vocabulary(
                url,
                "contributorsroles",
                "Contributors Roles",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "descriptiontypes": _generate_rdm_vocabulary(
                url,
                "descriptiontypes",
                "Description Types",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "datetypes": _generate_rdm_vocabulary(
                url,
                "datetypes",
                "Date Types",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "relationtypes": _generate_rdm_vocabulary(
                url,
                "relationtypes",
                "Relation Types",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
            "removalreasons": _generate_rdm_vocabulary(
                url,
                "removalreasons",
                "Removal Reasons",
                special=False,
                schema="local://vocabulary-v1.0.0.json",
            ),
        },
    )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base utilities for the commandline client."""

import functools
import inspect
import json
from collections.abc import Callable, Generator, Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Protocol, Self, get_origin

import yaml
from rich.console import Console
from rich.table import Table

from nrp_cmd.config import Config
from nrp_cmd.converter import converter

from .arguments import OutputFormat

if TYPE_CHECKING:
    from io import TextIOWrapper


def format_output(output_format: OutputFormat, data: Any) -> str:  # noqa: ANN401
    """Format the output according to the given format.

    Only the json, jsonl and yaml formats are supported.
    """
    if not isinstance(data, (dict, list)):
        data = converter.unstructure(data)
    if isinstance(data, list):
        unstructured_data = []
        for record in data:
            if not isinstance(record, (dict, list)):
                unstructured_data.append(converter.unstructure(record))
            else:
                unstructured_data.append(record)
        data = unstructured_data
    match output_format:
        case OutputFormat.JSON:
            return json.dumps(data, indent=4)
        case OutputFormat.JSON_LINES:
            return json.dumps(
                data, ensure_ascii=False, separators=(",", ":"), indent=None
            ).replace("\n", "\\n")
        case OutputFormat.YAML:
            return yaml.safe_dump(data)
        case _:
            raise ValueError(f"Unknown output format: {output_format}")


def format_table_value(value: list | dict | Any) -> str:  # noqa: ANN401
    """Format a value (dist, list or primitive) for a table."""
    if isinstance(value, list):
        if all(isinstance(x, dict) for x in value):
            return ",\n".join(
                json.dumps(x, indent=2, ensure_ascii=False) for x in value
            )
        return ", ".join(str(x) for x in value)
    if isinstance(value, dict):
        return json.dumps(value, indent=2, ensure_ascii=False)
    return str(value)


def write_table_row(table: Table, k: str, v: Any, prefix: str = "") -> None:  # noqa: ANN401
    """Write a row into a table."""
    if isinstance(v, dict):
        table.add_row(prefix + k, "")
        for k1, v1 in v.items():
            table.add_row(prefix + "    " + k1, format_table_value(v1))
    else:
        table.add_row(prefix + k, format_table_value(v))


def _is_typer_argument(name: str, parameter: inspect.Parameter) -> bool:
    """Check if the annotation contains typer argument / option."""
    annotation = parameter.annotation
    if get_origin(annotation) is not Annotated:
        return False

    for md in annotation.__metadata__:
        if isinstance(md, (typer.models.ArgumentInfo, typer.models.OptionInfo)):
            return True
    return False


def async_command(func: Any) -> Callable[..., None]:
    """Run an async function synchronously.

    This decorator is used to run an async function synchronously.
    """
    import asyncio

    import uvloop

    @functools.wraps(func)
    def runner(*args: Any, **kwargs: Any):
        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            runner.run(func(*args, **kwargs))

    return runner


def set_variable(config: Config, name: str, value: list[str] | str) -> None:
    """Set a variable into the configuration (local directory or global)."""
    if not isinstance(value, list):
        value = [value]

    variables = config.load_variables()

    if name.startswith("@"):
        name = name[1:]
    if name.startswith("+"):
        name = name[1:]
        variables[name] = variables.get(name, []) + value  # type: ignore
    else:
        variables[name] = value
    variables.save()


class TableMaker(Protocol):
    """A function that converts data into a table or a list of tables."""

    def __call__(
        self, data: Any, **kwargs: Any
    ) -> (
        Table
        | Generator[Table, None, None]
        | list[Table]
        | str
        | Generator[str, None, None]
    ):
        """Create a table or a list of tables from the data."""
        ...


class OutputWriter:
    """A Writer capable of writing output to a file or console and formatting it to table, json or yaml."""

    def __init__(
        self,
        output_file: Path | None,
        output_format: OutputFormat | None,
        console: Console,
        table_maker: TableMaker,
    ) -> None:
        """Initialize the output writer."""
        self.output_file = output_file
        self.output_format = output_format
        self.console = console
        self.table_maker = table_maker
        self._stream: TextIOWrapper | None = None
        self._multiple = False
        self._output_encountered = False

    def __enter__(self) -> "Self":
        """Enter the context manager, opening the output file if provided."""
        if self.output_file:
            self._stream = self.output_file.open("w")
        else:
            self._stream = self.console
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # noqa
        if self._multiple and self.output_format == OutputFormat.JSON:
            self._print("]")
        if self.output_file and self._stream:
            self._stream.close()
            self._stream = None

    def output(self, data: Any) -> None:  # noqa: ANN401
        """Output the data.

        If the output_format is 'table', the data will be passed to the table_maker function
        and printed as a table. Otherwise, the data will be formatted according to the output_format
        and printed.
        """
        assert (
            self._stream is not None
        ), "Output stream is not initialized, please use the context manager."

        if self._multiple and self._output_encountered:
            match self.output_format:
                case OutputFormat.JSON:
                    self._print(",")
                case OutputFormat.JSON_LINES:
                    self._print()
                case OutputFormat.YAML:
                    self._print("\n---")
                case _:
                    pass

        if self.output_format == OutputFormat.TABLE or self.output_format is None:
            if not isinstance(self._stream, Console):
                console = Console(file=self._stream)
            else:
                console = self._stream

            tables: Iterable[Table]
            if inspect.isgeneratorfunction(self.table_maker):
                tables = list(self.table_maker(data))
            else:
                generated_tables: Table | list[Table] = self.table_maker(data=data)  # type: ignore
                if not isinstance(generated_tables, (tuple, list)):
                    tables = [generated_tables]
                else:
                    tables = generated_tables

            for table in tables:
                console.print(table)
        else:
            data = format_output(self.output_format, data)
            self._print(data, end="" if self._multiple else "\n")
        self._output_encountered = True

    def multiple(self) -> None:
        """Set the output to be multiple records."""
        self._multiple = True
        if self.output_format == OutputFormat.JSON:
            assert (
                self._stream is not None
            ), "Output stream is not initialized, please use the context manager."
            self._print("[")

    def _print(self, data: Any = None, end: str = "\n") -> None:  # noqa: ANN401
        """Print the data to the output stream."""
        assert (
            self._stream is not None
        ), "Output stream is not initialized, please use the context manager."
        if hasattr(self._stream, "print"):
            self._stream.print(data, end=end)
        else:
            if data is not None:
                self._stream.write(str(data))
            self._stream.write(end)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Contains the limiter class for limiting the number of simultaneous connections.

Note: the limiter is not propagated to created threads - each thread has by default
its own limiter with the default capacity of 10 connections.

Use LimitedThread to create a thread with the same limiter as the parent thread.
"""

import contextlib
import contextvars
from collections.abc import Generator
from threading import Lock, Semaphore, Thread

from yarl import URL

limiter_lock = Lock()


class Limiter(Semaphore):
    """A class to limit the number of simultaneous connections."""

    def __init__(self, capacity: int, per_server_capacity: int = -1):
        """Initialize the limiter.

        :param capacity:    the number of simultaneous connections
        """
        if capacity <= 0:
            capacity = 10
        self.capacity = capacity
        super().__init__(capacity)

    @property
    def free(self) -> int:
        """The number of free slots.

        :return:   the number of remaining connections (approximate)
        """
        return self._value

    @contextlib.contextmanager
    def limit(self, url: URL):
        with self:
            yield

current_limiter_var = contextvars.ContextVar[Limiter](
    "current_limiter", default=Limiter(10)
)

class CurrentLimiterProxy:
    """A class to manage the current limiter."""

    @property
    def free(self) -> int:
        """The number of free slots.

        :return:   the number of remaining connections (approximate)
        """
        return current_limiter_var.get().free

    @contextlib.contextmanager
    def limit(self, url: URL):
        with current_limiter_var.get().limit(url):
            yield


current_limiter = CurrentLimiterProxy()

@contextlib.contextmanager
def limit_connections(
    max_connections: int, per_server_connections: int = -1
) -> Generator[None, None, None]:
    """Limit the number of connections called in the context of the with statement."""
    token = current_limiter_var.set(Limiter(max_connections, per_server_connections))
    try:
        yield
    finally:
        current_limiter_var.reset(token)


class LimitedThread(Thread):
    """A thread with a limited number of connections."""

    def __init__(self, *args, **kwargs):
        self.ctx = contextvars.copy_context()
        super().__init__(*args, **kwargs)

    def run(self):
        # This code runs in the target, child class:
        self.ctx.run(super().run)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Asynchronous task group."""

import asyncio
from collections.abc import Callable, Coroutine
from contextvars import Context
from typing import Any


class TaskGroup(asyncio.TaskGroup):
    """Compatibility layer for TaskGroup.

    Using factory for coro so that we can have a similar API to the sync version.
    """

    def create_task[_T](  # type: ignore
        self,
        coro: Callable[[], Coroutine[Any, Any, _T]],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> asyncio.Task[_T]:
        return super().create_task(coro(), name=name, context=context)

Task = asyncio.Task

__all__ = ("TaskGroup", "Task")


#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Configuration of the repository and repository access classes."""

from typing import Optional

from attrs import define
from yarl import URL

from ..types.info import RepositoryInfo


@define(kw_only=True)
class RepositoryConfig:
    """Configuration of the repository."""

    alias: str
    """The local alias of the repository."""

    url: URL
    """The api URL of the repository, usually something like https://repository.org/api."""

    token: Optional[str] = None
    """Bearer token"""

    verify_tls: bool = True
    """Verify the TLS certificate in https"""

    retry_count: int = 10
    """Number of times idempotent operations will be retried if something goes wrong."""

    retry_after_seconds: int = 10
    """If server does not suggest anything else, retry after this interval in seconds"""

    info: RepositoryInfo | None = None
    """Cached repository info"""

    enabled: bool = True
    
    class Config:  # noqa
        extra = "forbid"

    @property
    def well_known_repository_url(self) -> URL:
        """Return URL to the well-known repository endpoint."""
        return self.url / ".well-known" / "repository/"

    def search_url(self, model: str | None) -> URL:
        """Return URL to search for published records within a model."""
        assert self.info is not None
        model = model or self._default_model_name
        if model:
            return self.info.models[model].links.published
        return self.info.links.records

    def user_search_url(self, model: str | None) -> URL:
        """Return URL to search for user's records within a model."""
        assert self.info is not None
        model = model or self._default_model_name
        if model:
            user_records = self.info.models[model].links.user_records
            if user_records is None:
                return self.search_url(model)
            return user_records
        return self.info.links.user_records

    def create_url(self, model: str | None) -> URL:
        """Return URL to create a new record within a model."""
        assert self.info is not None
        model = model or self._default_model_name
        if model:
            return self.info.models[model].links.api
        return self.info.links.records

    def read_url(self, model: str | None, record_id: str | URL) -> URL:
        """Return URL to a published record within a model."""
        assert self.info is not None
        if isinstance(record_id, URL):
            return record_id
        if record_id.startswith("https://"):
            return URL(record_id)
        model = model or self._default_model_name
        if model:
            return self.info.models[model].links.api / record_id
        return self.info.links.records / record_id

    def user_read_url(self, model: str | None, record_id: str | URL) -> URL:
        """Return URL to a draft record within a model."""
        assert self.info is not None
        if isinstance(record_id, URL):
            return record_id
        if record_id.startswith("https://"):
            return URL(record_id)
        model = model or self._default_model_name
        if model:
            return self.info.models[model].links.api / record_id / "draft"
        return self.info.links.records / record_id / "draft"

    @property
    def requests_url(self) -> URL:
        """Return URL to the requests endpoint."""
        assert self.info is not None
        return self.info.links.requests

    @property
    def _default_model_name(self) -> str | None:
        """Return the default model name if there is only one model in the repository."""
        if self.info and len(self.info.models) == 1:
            return next(iter(self.info.models))
        return None



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command line interface for getting records."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import rich_click as click
from rich.console import Console

from nrp_cmd.async_client import (
    AsyncRepositoryClient,
    limit_connections,
)
from nrp_cmd.async_client.connection import limit_connections
from nrp_cmd.cli.base import OutputWriter, async_command
from nrp_cmd.cli.records.get import read_record
from nrp_cmd.config import Config
from nrp_cmd.types.records import Record
from nrp_cmd.types.requests import Request

from ..arguments import (
    Model,
    Output,
    argument_with_help,
    with_config,
    with_model,
    with_output,
    with_repository,
    with_resolved_vars,
    with_verbosity,
)
from .table_formatter import format_request_table

if TYPE_CHECKING:
    from nrp_cmd.types.requests import RequestType


@with_config
@with_repository
@with_resolved_vars("record_id")
@with_output
@with_model
@with_verbosity
@argument_with_help("request_type_id", type=str, help="Request type ID")
@argument_with_help("record_id", type=str, help="Record ID")
@argument_with_help("variable", type=str, required=False, help="Variable name")
@click.option("--submit/--no-submit", default=True, help="Submit the request")
@async_command
async def create_request(
    *,
    config: Config,
    repository: Optional[str] = None,
    request_type_id: str,
    record_id: str,
    variable: Optional[str] = None,
    out: Output,
    model: Model,
    submit: bool = True,
) -> None:
    """Create a request."""
    console = Console()

    with limit_connections(10):
        (
            record,
            _final_record_id,
            _repository_config,
            _record_client,
            repository_client,
        ) = await read_record(
            record_id,
            repository,
            config,
            False,
            model.model,
            model.published,
            model.draft,
        )
        request = await create_request_helper(
            console, repository_client, record, request_type_id, submit
        )

    with OutputWriter(
        out.output, out.output_format, console, format_request_table
    ) as printer:
        printer.output(request)

    if variable:
        variables = config.load_variables()
        variables[variable[1:]] = [str(request.links.self_)]
        variables.save()


async def create_request_helper(
    console: Console,
    repository_client: AsyncRepositoryClient,
    record: Record,
    request_type_id: str,
    submit: bool,
) -> Request:
    request_types = await repository_client.requests.applicable_requests(record)

    request_type: RequestType | None = next(
        (rt for rt in request_types.hits if rt.type_id == request_type_id), None
    )
    if not request_type:
        console.print(
            f"[red]Request type {request_type_id} not applicable to record {record.id}[/red]"
        )
        raise click.Abort()

    return await repository_client.requests.create(request_type, {}, submit=submit)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Memory sink and source."""


import asyncio
import base64
import hashlib

from .base import DataSink, DataSource, InputStream, OutputStream, SinkState


class MemorySink(DataSink):
    """Implementation of a sink that writes data to memory."""

    def __init__(self):
        """Initialize the sink."""
        self._state = SinkState.NOT_ALLOCATED
        self._buffer = None

    async def allocate(self, size: int) -> None:
        """Allocate space for the sink."""
        self._buffer = bytearray(size)
        self._state = SinkState.ALLOCATED

    async def open_chunk(self, offset: int = 0) -> OutputStream:  # type: ignore
        """Open a chunk of the sink for writing."""
        if self._state != SinkState.ALLOCATED:
            raise RuntimeError("Sink not allocated")

        return MemoryWriter(self._buffer, offset)  # noqa

    async def close(self) -> None:
        """Close the sink."""
        self._state = SinkState.CLOSED

    @property
    def state(self) -> SinkState:
        """Return the state of the sink."""
        return self._state


class MemorySource(DataSource):
    """A data source that reads data from memory."""

    has_range_support = True

    def __init__(self, data: bytes, content_type: str):
        """Initialize the data source.

        :param data:                the data to be read
        :param content_type:        the content type of the data
        """
        self._data = data
        self._content_type = content_type

    async def open(
        self, offset: int = 0, count: int | None = None
    ) -> InputStream:
        """Open the data source for reading."""
        if count is not None:
            return MemoryReader(self._data[offset : offset + count])
        else:
            return MemoryReader(self._data[offset:])

    async def size(self) -> int: 
        """Return the size of the data."""
        return len(self._data)

    async def content_type(self) -> str:
        """Return the content type of the data."""
        return self._content_type

    async def close(self) -> None:
        """Close the data source."""
        pass

    async def checksum(
        self, algo: str = "md5", offset: int = 0, count: int | None = None
    ) -> str:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._checksum, algo, offset, count)

    def _checksum(self, algo: str, offset: int, count: int | None) -> str:
        """Calculate the checksum of the data."""
        if count is not None:
            data = self._data[offset : offset + count]
        else:
            data = self._data[offset:]

        hasher = hashlib.new(algo)
        hasher.update(data)
        return base64.b64encode(hasher.digest()).decode("ascii")

    def supported_checksums(self) -> list[str]:
        """Return a list of supported checksum algorithms."""
        return list(hashlib.algorithms_available)


class MemoryReader(InputStream):
    """A reader for in-memory data."""

    def __init__(self, data: bytes):
        """Initialize the reader.

        :param data:        the data that will be read
        """
        self._data = data

    def __len__(self):
        """Return the length of the data."""
        return len(self._data) if self._data is not None else 0

    def __bool__(self):
        """Return whether there is data to read."""
        return bool(self._data)

    def __aiter__(self):
        """We are our own iterator."""
        return self

    async def __anext__(self):
        """Simulate normal file iteration."""
        if self._data:
            ret = self._data
            self._data = None
            return ret
        else:
            raise StopAsyncIteration

    async def read(self, size: int = -1) -> bytes:
        """Read data from the buffer.

        :param size: the number of bytes to read
        :return: the data read
        """
        if self._data is None:
            return b""
        
        if size < 0:
            return await anext(self)

        size = min(size, len(self._data))
        ret = self._data[:size]
        self._data = self._data[size:]
        return ret
    
    async def close(self) -> None:
        """Close the reader."""
        pass

class MemoryWriter(OutputStream):
    """Implementation of a writer that writes data to memory."""

    def __init__(self, buffer: bytearray, offset: int):
        """Initialize the writer.

        :param buffer: The buffer where the data will be written.
        :param offset: The offset in bytes from the start of the buffer.
        """
        self._buffer = buffer
        self._offset = offset

    async def write(self, b: bytes) -> int:
        """Write data to the buffer.

        :param b: the bytes to be written
        :return:  number of bytes written
        """
        self._buffer[self._offset : self._offset + len(b)] = b
        self._offset += len(b)
        return len(b)

    async def close(self) -> None:
        """Close the writer."""
        pass



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# invenio-nrp is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Helper functions for working with metadata on commandline."""

import json
import sys
from pathlib import Path
from typing import Any


class MetadataIsNotJSON(Exception):
    """Raised when metadata is not a valid JSON object."""


def read_metadata(metadata: str) -> dict[str, Any] | list[dict[str, Any]]:
    """Read metadata from a string.

    If the string is '-', read metadata from stdin.
    If the string is a path to a file, read metadata from the file.
    If the metadata look like a JSON object, parse it as JSON.
    Otherwise raises an MetadataIsNotJSON exception.
    """
    if metadata == "-":
        # read metadata from stdin
        metadata = sys.stdin.read()

    metadata = metadata.strip()
    if not (metadata.startswith("{") or metadata.startswith("[")):
        pth = Path(metadata)
        if pth.exists():
            # metadata is path on filesystem
            metadata = pth.read_text()
        else:
            raise MetadataIsNotJSON(
                "Metadata must be a stringified JSON object or path to an existing file."
            )

    return json.loads(metadata)



import importlib.metadata as importlib_metadata
from functools import lru_cache

from yarl import URL

from ..config import Config, RepositoryConfig
from . import base_client, connection, doi, invenio, streams
from .base_client import RecordStatus, SyncRepositoryClient
from .doi import resolve_doi


@lru_cache(maxsize=1)
def sync_client_classes() -> list[type[SyncRepositoryClient]]:
    """Load all available synchronous client classes."""
    sync_client_classes: list[type[SyncRepositoryClient]] = []
    for ep in importlib_metadata.entry_points(group="nrp_cmd.sync_client"):
        sync_client_classes.append(ep.load())
    return sync_client_classes


def get_sync_client(
    repository: str | URL,
    refresh: bool = False,
    config: Config | None = None,
) -> SyncRepositoryClient:
    """Get a synchronous client for the given repository.

    :param repository: the repository alias or URL
    :param refresh: whether to refresh the client configuration from the server
    :param max_connections: the maximum number of parallel connections
    :param config: the configuration to use. If not given, the configuration is loaded
    from the configuration file.
    :return: a synchronous client for the repository
    """
    if not config:
        config = Config.from_file()

    repository_config = config.find_repository(repository)
    for sync_client_class in sync_client_classes():
        if sync_client_class.can_handle_repository(
            repository_config.url, verify_tls=repository_config.verify_tls
        ):
            return sync_client_class.from_configuration(
                repository_config, refresh=refresh
            )
    raise ValueError(f"No sync client found for repository {repository_config.url}")


def get_repository_from_record_id(
    connection: connection.SyncConnection,
    record_id: str,
    config: Config,
    repository: str | None = None,
) -> tuple[str, RepositoryConfig]:
    """Try to get a repository from the record id.

    :param record_id: The record id (might be id, url, doi)
    :param config: The configuration of known repositories
    :param repository: The optional repository alias to use. If not passed in, the call will try to
                          resolve the repository from the record id.
    """
    if record_id.startswith("doi:"):
        record_id = resolve_doi(connection, record_id[4:])
    elif record_id.startswith("https://doi.org/"):
        record_id = resolve_doi(connection, record_id[len("https://doi.org/") :])

    if repository:
        repository_config = config.get_repository(repository)
        return record_id, repository_config

    if not record_id.startswith("https://"):
        return record_id, config.default_repository

    repository_config = config.get_repository_from_url(record_id)

    # if it is an api path, return it as it is
    record_url = URL(record_id)
    if record_url.path.startswith("/api/"):
        return str(record_id), repository_config

    # try to head the record to get the id
    api_url = None
    resp = connection.head(url=record_url, get_links=True)
    for linkset in resp.getall("linkset"):
        api_url = linkset["url"]
        break

    if api_url:
        return str(api_url), repository_config
    else:
        return str(record_url), repository_config


__all__ = (
    "get_sync_client",
    "SyncRepositoryClient",
    "connection",
    "invenio",
    "streams",
    "base_client",
    "doi",
    "get_repository_from_record_id",
    "RecordStatus",
)
